from fiscalsim_us.model_api import *


class is_married(Variable):
    value_type = bool
    entity = Family
    label = "Married"
    documentation = "Whether the adults in this family are married."
    definition_period = YEAR

    def formula(family, period, parameters):
        # If any tax unit is a married filer, assume the family is.
        person = family.members
        filing_status = person.tax_unit("filing_status", period)
        filing_status_type = filing_status.possible_values
        person_is_married = is_in(
            person.tax_unit("filing_status", period),
            [
                filing_status_type.JOINT,
                filing_status_type.SEPARATE,
            ],
        )
        return family.any(person_is_married)



from fiscalsim_us.model_api import *
import numpy as np


class ar_income_tax_before_non_refundable_credits(Variable):
    "Line 29 of form AR1000F"
    value_type = float
    entity = TaxUnit
    label = "Arkansas income tax before non refundable credits"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_Final_AR1000ES.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.rates
        taxable_income = tax_unit("ar_taxable_income", period)
        high_income_threshold = p.regular_bracket_max
        litc = tax_unit("ar_low_income_credit", period)
        high_income_reduction = tax_unit("ar_high_income_reduction", period)

        def round_to_nearest_50(num):
            # Calculate the nearest multiple of 100
            nearest_multiple_of_100 = np.round(num / 100, 0) * 100

            # Get the last two digits
            last_two_digits = num % 100

            # Determine the closest ending in "50"
            if np.isscalar(num):
                if last_two_digits <= 50 and last_two_digits >= 1:
                    rounded_income = nearest_multiple_of_100 + 50
                    return rounded_income
                else:
                    rounded_income = nearest_multiple_of_100 - 50
                    return rounded_income
            else:
                rounded_income = np.zeros_like(num)
                rounded_income[
                    (last_two_digits <= 50) & (last_two_digits >= 1)
                ] = (
                    nearest_multiple_of_100[
                        (last_two_digits <= 50) & (last_two_digits >= 1)
                    ]
                    + 50
                )
                rounded_income[
                    (last_two_digits > 50) | (last_two_digits < 1)
                ] = (
                    nearest_multiple_of_100[
                        (last_two_digits > 50) | (last_two_digits < 1)
                    ]
                    - 50
                )
                return rounded_income

        rounded_taxable_income = round_to_nearest_50(taxable_income)

        if np.isscalar(rounded_taxable_income):
            tax = np.round(
                where(
                    taxable_income <= high_income_threshold,
                    p.rates.calc(rounded_taxable_income),
                    p.high_income_rates.calc(rounded_taxable_income),
                )
                - litc
                - high_income_reduction,
                0,
            )
        else:
            tax = np.zeros_like(rounded_taxable_income)
            tax[taxable_income <= high_income_threshold] = np.round(
                p.rates.calc(
                    rounded_taxable_income[
                        taxable_income <= high_income_threshold
                    ]
                )
                - litc[taxable_income <= high_income_threshold]
                - high_income_reduction[
                    taxable_income <= high_income_threshold
                ],
                0,
            )
            tax[taxable_income > high_income_threshold] = np.round(
                p.high_income_rates.calc(
                    rounded_taxable_income[
                        taxable_income > high_income_threshold
                    ]
                )
                - litc[taxable_income > high_income_threshold]
                - high_income_reduction[
                    taxable_income > high_income_threshold
                ],
                0,
            )

        lump_sum_dist_tax = tax_unit("ar_lump_sum_dist_tax", period)

        # Additional tax on IRA and qualified plan withdrawal and overpayment, not modelled

        total_tax = tax + lump_sum_dist_tax

        return total_tax



from fiscalsim_us.model_api import *


class mt_light_vehicle_registration_deduction(Variable):
    """
    Line 6 on itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana light vehicle registration fees"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class va_additions_to_federal_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA additions to federal agi https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-adj-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        income_from_obligations_fed_exempt = tax_unit(
            "va_income_from_obligations_fed_exempt", period
        )

        fixed_date_conformity_additions = tax_unit(
            "va_fixed_date_conformity_additions", period
        )

        additions_to_fed_agi = (
            income_from_obligations_fed_exempt
            + fixed_date_conformity_additions
        )

        return additions_to_fed_agi



import pytest
from fiscalsim_us import CountryTaxBenefitSystem
from policyengine_core.simulations import SimulationBuilder

DEFAULT_SITUATION = {
    "people": {"person": {}},
    "tax_units": {"tax_unit": {"members": ["person"]}},
    "families": {"family": {"members": ["person"]}},
    "spm_units": {"spm_units": {"members": ["person"]}},
    "households": {"household": {"members": ["person"]}},
}

EXEMPTIONS = (
    "household_income_decile",  # because DEFAULT_SITUATION has no weights
    "spm_unit_income_decile",  # because DEFAULT_SITUATION has no weights
    "income_decile",  # because DEFAULT_SITUATION has no weights
    "tanf_max_amount",
    "tanf",
    "tanf_amount_if_eligible",
    "tanf_countable_income",
    "tanf_initial_employment_deduction",
    "is_tanf_initial_eligible",
    "is_tanf_continuous_eligible",
    "is_tanf_eligible",
    "snap_emergency_allotment_monthly",
)


system = CountryTaxBenefitSystem()

simulation = SimulationBuilder().build_from_dict(system, DEFAULT_SITUATION)


@pytest.mark.parametrize("variable", system.variables)
def test_variable(variable: str) -> None:
    if variable not in EXEMPTIONS:
        simulation.calculate(variable, 2022)



from fiscalsim_us.model_api import *


class ar_additional_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas additional tax credit for qualified individuals"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)

        income = tax_unit("ar_taxable_income", period)
        p = parameters(period).gov.states.ar.tax.income.credits.other_credits

        amount = p.additional_amount

        joint_multiple = p.additional_credit_joint_multiple

        credit = amount.calc(income)

        credit = where(
            filing_status == filing_status.possible_values.JOINT,
            credit * joint_multiple,
            credit,
        )

        return credit



import numpy as np
from fiscalsim_us.model_api import *


class is_older_child_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Older children"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#l_1_D"

    def formula(person, period, parameters):
        age = person("age", period)
        ma = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.older_child
        income = person("medicaid_income_level", period)
        is_older_child = ma.age_range.calc(age)
        state = person.household("state_code_str", period)
        income_limit = ma.income_limit[state]

        try:
            result = is_older_child & (income < income_limit)
            return result
        except Exception as e:
            print(f"Calculation failed: {str(e)}")
            print(
                f"is_older_child: type={type(is_older_child)}, value={is_older_child}"
            )
            print(f"income: type={type(income)}, value={income}")
            print(
                f"income_limit: type={type(income_limit)}, value={income_limit}"
            )

            # Convert to numpy arrays and ensure correct types
            is_older_child = np.asarray(
                is_older_child if is_older_child is not None else False
            ).astype(bool)
            income = np.asarray(income if income is not None else 0).astype(
                float
            )
            income_limit = np.asarray(
                income_limit if income_limit is not None else 0
            ).astype(float)

            # Attempt calculation again
            result = is_older_child & (income < income_limit)
            print("Calculation succeeded after type conversion")
            return result



from fiscalsim_us.model_api import *


class mn_casualty_theft_deduction(Variable):
    """
    Minnesota casulaty/ theft deduction
    Line 19 of 2023 M1SA, Minnesota Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "MN charitable donation deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.deductions
        fed_agi = tax_unit("adjusted_gross_income", period)
        loss = add(tax_unit, period, ["casualty_loss"])

        adj_loss = max_(0, loss - p.theft_deduction_floor)

        return max_(0, adj_loss - fed_agi * p.theft_deduction_income_mult)



from fiscalsim_us.model_api import *


class exemption_phase_out_start(Variable):
    value_type = float
    entity = TaxUnit
    label = "Exemption phase-out start"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return parameters(period).gov.irs.income.exemption.phase_out.start[
            tax_unit("filing_status", period)
        ]



from fiscalsim_us.model_api import *


class pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Pension contributions"
    unit = USD
    documentation = "Contributions to IRAs, SEP, and other pension funds."
    definition_period = YEAR

    adds = ["ira_contributions", "sep_simple_qualified_plan_contributions"]



from fiscalsim_us.model_api import *


class dwks14(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS14"
    unit = USD
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )

    def formula(tax_unit, period, parameters):
        dwks1 = tax_unit("taxable_income", period)
        dwks13 = tax_unit("dwks13", period)
        return max_(0, dwks1 - dwks13) * tax_unit("hasqdivltcg", period)



from fiscalsim_us.model_api import *


class statelocal_sales_or_prior_inctax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Prior year state and local income taxes paid or state and local sales taxes paid"
    unit = USD

    def formula(tax_unit, period, parameters):
        # Only sales or income tax can be itemized, but not both.
        prior_year_statelocal_inctax = add(
            tax_unit,
            period,
            [
                "prior_year_state_income_tax_paid",
                "prior_year_local_income_tax_paid",
            ],
        )
        sales_tax = add(
            tax_unit, period, ["state_sales_tax", "local_sales_tax"]
        )
        return max_(prior_year_statelocal_inctax, sales_tax)



from fiscalsim_us.model_api import *


class ut_total_income(Variable):
    """
    Line 6 on Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT total income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    adds = ["adjusted_gross_income", "ut_additions_to_income"]



from fiscalsim_us.model_api import *


class ar_metabolic_disorder_credit(Variable):
    "Line 4 of form AR1000TC"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Phenylketonuria Disorder Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TC_Schedule_of_TaxCredits_andBusiness_IncentiveCredits.pdf"
    defined_for = StateCode.AR


def formula(tax_unit, period, parameters):
    p = parameters(period).gov.states.ar.income.credits.other_credits

    maximum_credit = p.metabolic

    expenses = tax_unit("ar_metabolic_disorder_expenses", period)

    carryover = tax_unit("ar_metabolic_credit_carryover", period)

    total_available = expenses + carryover

    total_allowable_credit = min_(maximum_credit, total_available)

    tax_due = tax_unit("ar_income_tax_before_non_refundable_credits", period)
    -tax_unit("ar_personal_credits", period) - tax_unit(
        "ar_political_contribution_credit", period
    )
    (
        -tax_unit("ar_other_state_credit", period)
        - tax_unit("ar_adoption_expense_credit", period)
        - tax_unit("ar_stillborn_child_credit")
    )
    (
        -tax_unit("ar_additional_tax_credit")
        - tax_unit("ar_inflationary_relief_credit")
        - tax_unit("ar_cdcc", period)
    )

    credit = min_(total_allowable_credit, tax_due)

    return credit



from fiscalsim_us.model_api import *


class ar_adoption_expense_credit(Variable):
    "Line 3 of form AR1000TC"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Adoption Expense Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TC_Schedule_of_TaxCredits_andBusiness_IncentiveCredits.pdf"
    defined_for = StateCode.AR


def formula(tax_unit, period, parameters):
    p = parameters(period).gov.states.ar.income.credits.other_credits

    match = p.adoption_credit

    return match * tax_unit("qualified_adoption_expenses", period)



from fiscalsim_us.model_api import *


class StateLivingArrangement(Enum):
    FULL_COST = "Full cost of living"
    SHARED_EXPENSES = "Shared expenses"
    HOUSEHOLD_OF_ANOTHER = "Household of another"
    REST_HOME = "Rest home"
    MEDICAID_FACILITY = "Medicaid facility"
    ASSISTED_LIVING = "Assisted living"


class state_living_arrangement(Variable):
    value_type = Enum
    entity = Household
    label = "Federal living arrangement"
    definition_period = YEAR
    possible_values = StateLivingArrangement
    default_value = StateLivingArrangement.FULL_COST



from fiscalsim_us.model_api import *


class num(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "2 when filing_status is married filing jointly; otherwise 1"
    )
    unit = USD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        return where(
            filing_status == filing_status.possible_values.JOINT, 2, 1
        )



from fiscalsim_us.model_api import *


class snap_min_allotment(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "SNAP minimum allotment"
    documentation = (
        "Minimum allotment for SNAP based on household size and state"
    )
    unit = USD

    def formula(spm_unit, period, parameters):
        # Parameters for the minimum benefit.
        snap = parameters(period).gov.usda.snap
        min_allotment = snap.min_allotment
        # Calculate the relevant maximum benefit, defined as the maximum
        # benefit for a household of a certain size in their state.
        snap_region = spm_unit.household("snap_region_str", period)
        relevant_max_allotment = snap.max_allotment.main[snap_region][
            str(min_allotment.relevant_max_allotment_household_size)
        ]
        # Minimum benefits only apply to households up to a certain size.
        size = spm_unit("spm_unit_size", period)
        eligible = size <= min_allotment.maximum_household_size
        return eligible * min_allotment.rate * relevant_max_allotment



import logging
from policyengine_core.data import Dataset
import h5py
from fiscalsim_us.data.datasets.cps.raw_cps import (
    RawCPS_2020,
    RawCPS_2021,
    RawCPS_2022,
    RawCPS,
)
from fiscalsim_us.data.datasets.cps.uprated_cps import UpratedCPS
from fiscalsim_us.data.storage import STORAGE_FOLDER
from pandas import DataFrame, Series
import numpy as np
import pandas as pd
import os
import yaml
from typing import Type


class CPS(Dataset):
    name = "cps"
    label = "CPS"
    raw_cps: Type[RawCPS] = None
    data_format = Dataset.ARRAYS

    def generate(self):
        """Generates the Current Population Survey dataset for PolicyEngine US microsimulations.
        Technical documentation and codebook here: https://www2.census.gov/programs-surveys/cps/techdocs/cpsmar21.pdf
        """

        raw_data = self.raw_cps().load()
        cps = h5py.File(self.file_path, mode="w")

        ENTITIES = ("person", "tax_unit", "family", "spm_unit", "household")
        person, tax_unit, family, spm_unit, household = [
            raw_data[entity] for entity in ENTITIES
        ]

        add_id_variables(cps, person, tax_unit, family, spm_unit, household)
        add_personal_variables(cps, person)
        add_personal_income_variables(cps, person, self.raw_cps.time_period)
        add_spm_variables(cps, spm_unit)
        add_household_variables(cps, household)

        raw_data.close()
        cps.close()

        cps = h5py.File(self.file_path, mode="a")
        add_silver_plan_cost(self, cps, 2022)
        cps.close()


def add_silver_plan_cost(self, cps: h5py.File, year: int):
    """Adds the second-lowest silver plan cost for each tax unit, based on geography.

    Args:
        cps (h5py.File): The CPS dataset file.
        year (int): The year of the data.
    """
    from fiscalsim_us import Microsimulation

    sim = Microsimulation(dataset=self)
    slspc = sim.calc("second_lowest_silver_plan_cost", year).values

    cps["second_lowest_silver_plan_cost"] = slspc


def add_id_variables(
    cps: h5py.File,
    person: DataFrame,
    tax_unit: DataFrame,
    family: DataFrame,
    spm_unit: DataFrame,
    household: DataFrame,
) -> None:
    """Add basic ID and weight variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The person table of the ASEC.
        tax_unit (DataFrame): The tax unit table created from the person table
            of the ASEC.
        family (DataFrame): The family table of the ASEC.
        spm_unit (DataFrame): The SPM unit table created from the person table
            of the ASEC.
        household (DataFrame): The household table of the ASEC.
    """
    # Add primary and foreign keys
    cps["person_id"] = person.PH_SEQ * 100 + person.P_SEQ
    cps["family_id"] = family.FH_SEQ * 10 + family.FFPOS
    cps["household_id"] = household.H_SEQ
    cps["person_tax_unit_id"] = person.TAX_ID
    cps["person_spm_unit_id"] = person.SPM_ID
    cps["tax_unit_id"] = tax_unit.TAX_ID
    cps["spm_unit_id"] = spm_unit.SPM_ID
    cps["person_household_id"] = person.PH_SEQ
    cps["person_family_id"] = person.PH_SEQ * 10 + person.PF_SEQ

    # Add weights
    # Weights are multiplied by 100 to avoid decimals
    cps["person_weight"] = person.A_FNLWGT / 1e2
    cps["family_weight"] = family.FSUP_WGT / 1e2

    # Tax unit weight is the weight of the containing family.
    family_weight = Series(
        cps["family_weight"][...], index=cps["family_id"][...]
    )
    person_family_id = cps["person_family_id"][...]
    persons_family_weight = Series(family_weight[person_family_id])
    cps["tax_unit_weight"] = persons_family_weight.groupby(
        cps["person_tax_unit_id"][...]
    ).first()

    cps["spm_unit_weight"] = spm_unit.SPM_WEIGHT / 1e2

    cps["household_weight"] = household.HSUP_WGT / 1e2

    # Marital units

    marital_unit_id = person.PH_SEQ * 1e6 + np.maximum(
        person.A_LINENO, person.A_SPOUSE
    )

    # marital_unit_id is not the household ID, zero padded and followed
    # by the index within household (of each person, or their spouse if
    # one exists earlier in the survey).

    marital_unit_id = Series(marital_unit_id).rank(
        method="dense"
    )  # Simplify to a natural number sequence with repetitions [0, 1, 1, 2, 3, ...]

    cps["person_marital_unit_id"] = marital_unit_id.values
    cps["marital_unit_id"] = marital_unit_id.drop_duplicates().values


def add_personal_variables(cps: h5py.File, person: DataFrame) -> None:
    """Add personal demographic variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The CPS person table.
    """

    # The CPS provides age as follows:
    # 00-79 = 0-79 years of age
    # 80 = 80-84 years of age
    # 85 = 85+ years of age
    # We assign the 80 ages randomly between 80 and 84.
    # to avoid unrealistically bunching at 80.
    cps["age"] = np.where(
        person.A_AGE == 80,
        # NB: randint is inclusive of first argument, exclusive of second.
        np.random.randint(80, 85, len(person)),
        person.A_AGE,
    )
    # A_SEX is 1 -> male, 2 -> female.
    cps["is_female"] = person.A_SEX == 2
    # "Is...blind or does...have serious difficulty seeing even when Wearing
    #  glasses?" 1 -> Yes
    cps["is_blind"] = person.PEDISEYE == 1
    DISABILITY_FLAGS = [
        "PEDIS" + i for i in ["DRS", "EAR", "EYE", "OUT", "PHY", "REM"]
    ]
    cps["is_ssi_disabled"] = (person[DISABILITY_FLAGS] == 1).any(axis=1)

    def children_per_parent(col: str) -> pd.DataFrame:
        """Calculate number of children in the household using parental
            pointers.

        Args:
            col (str): Either PEPAR1 and PEPAR2, which correspond to A_LINENO
            of the person's first and second parent in the household,
            respectively.
        """
        return (
            person[person[col] > 0]
            .groupby(["PH_SEQ", col])
            .size()
            .reset_index()
            .rename(columns={col: "A_LINENO", 0: "children"})
        )

    # Aggregate to parent.
    res = (
        pd.concat(
            [children_per_parent("PEPAR1"), children_per_parent("PEPAR2")]
        )
        .groupby(["PH_SEQ", "A_LINENO"])
        .children.sum()
        .reset_index()
    )
    tmp = person[["PH_SEQ", "A_LINENO"]].merge(
        res, on=["PH_SEQ", "A_LINENO"], how="left"
    )
    cps["own_children_in_household"] = tmp.children.fillna(0)

    cps["has_marketplace_health_coverage"] = person.MRK == 1

    cps["cps_race"] = person.PRDTRACE
    cps["is_hispanic"] = person.PRDTHSP != 0

    cps["is_widowed"] = person.A_MARITL == 4
    cps["is_separated"] = person.A_MARITL == 6


def add_personal_income_variables(
    cps: h5py.File, person: DataFrame, year: int
):
    """Add income variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The CPS person table.
        year (int): The CPS year
    """
    # Get income imputation parameters.
    yamlfilename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "income_parameters.yaml"
    )
    with open(yamlfilename, "r", encoding="utf-8") as yamlfile:
        p = yaml.safe_load(yamlfile)
    assert isinstance(p, dict)

    # Assign CPS variables.
    cps["employment_income"] = person.WSAL_VAL
    cps["taxable_interest_income"] = person.INT_VAL * (
        p["taxable_interest_fraction"][year]
    )
    cps["tax_exempt_interest_income"] = person.INT_VAL * (
        1 - p["taxable_interest_fraction"][year]
    )
    cps["self_employment_income"] = person.SEMP_VAL
    cps["farm_income"] = person.FRSE_VAL
    cps["qualified_dividend_income"] = person.DIV_VAL * (
        p["qualified_dividend_fraction"][year]
    )
    cps["non_qualified_dividend_income"] = person.DIV_VAL * (
        1 - p["qualified_dividend_fraction"][year]
    )
    cps["rental_income"] = person.RNT_VAL
    # Assign Social Security retirement benefits if at least 62.
    MINIMUM_RETIREMENT_AGE = 62
    cps["social_security_retirement"] = np.where(
        person.A_AGE >= MINIMUM_RETIREMENT_AGE, person.SS_VAL, 0
    )
    # Otherwise assign them to Social Security disability benefits.
    cps["social_security_disability"] = (
        person.SS_VAL - cps["social_security_retirement"]
    )
    cps["unemployment_compensation"] = person.UC_VAL
    # Add pensions and annuities.
    cps_pensions = person.PNSN_VAL + person.ANN_VAL
    cps["taxable_pension_income"] = cps_pensions * (
        p["taxable_pension_fraction"][year]
    )
    cps["tax_exempt_pension_income"] = cps_pensions * (
        1 - p["taxable_pension_fraction"][year]
    )
    # Other income (OI_VAL) is a catch-all for all other income sources.
    # The code for alimony income is 20.
    cps["alimony_income"] = (person.OI_OFF == 20) * person.OI_VAL
    # The code for strike benefits is 12.
    cps["strike_benefits"] = (person.OI_OFF == 12) * person.OI_VAL
    cps["child_support_received"] = person.CSP_VAL
    # Assume all public assistance / welfare dollars (PAW_VAL) are TANF.
    # They could also include General Assistance.
    cps["tanf_reported"] = person.PAW_VAL
    cps["ssi_reported"] = person.SSI_VAL
    cps["pension_contributions"] = person.RETCB_VAL
    cps["long_term_capital_gains"] = person.CAP_VAL * (
        p["long_term_capgain_fraction"][year]
    )
    cps["short_term_capital_gains"] = person.CAP_VAL * (
        1 - p["long_term_capgain_fraction"][year]
    )
    cps["receives_wic"] = person.WICYN == 1
    cps["veterans_benefits"] = person.VET_VAL
    cps["workers_compensation"] = person.WC_VAL
    # Disability income has multiple sources and values split across two pairs
    # of variables. Include everything except for worker's compensation
    # (code 1), which is defined as WC_VAL.
    WORKERS_COMP_DISABILITY_CODE = 1
    disability_benefits_1 = person.DIS_VAL1 * (
        person.DIS_SC1 != WORKERS_COMP_DISABILITY_CODE
    )
    disability_benefits_2 = person.DIS_VAL2 * (
        person.DIS_SC2 != WORKERS_COMP_DISABILITY_CODE
    )
    cps["disability_benefits"] = disability_benefits_1 + disability_benefits_2
    # Expenses.
    # "What is the annual amount of child support paid?"
    cps["child_support_expense"] = person.CHSP_VAL
    cps["health_insurance_premiums"] = person.PHIP_VAL
    cps["medical_out_of_pocket_expenses"] = person.MOOP


def add_spm_variables(cps: h5py.File, spm_unit: DataFrame) -> None:
    SPM_RENAMES = dict(
        spm_unit_total_income_reported="SPM_TOTVAL",
        snap_reported="SPM_SNAPSUB",
        spm_unit_capped_housing_subsidy_reported="SPM_CAPHOUSESUB",
        free_school_meals_reported="SPM_SCHLUNCH",
        spm_unit_energy_subsidy_reported="SPM_ENGVAL",
        spm_unit_wic_reported="SPM_WICVAL",
        spm_unit_payroll_tax_reported="SPM_FICA",
        spm_unit_federal_tax_reported="SPM_FEDTAX",
        spm_unit_state_tax_reported="SPM_STTAX",
        spm_unit_work_childcare_expenses="SPM_CAPWKCCXPNS",
        spm_unit_medical_expenses="SPM_MEDXPNS",
        spm_unit_spm_threshold="SPM_POVTHRESHOLD",
        spm_unit_net_income_reported="SPM_RESOURCES",
        childcare_expenses="SPM_CHILDCAREXPNS",
    )

    for fiscalsim_variable, asec_variable in SPM_RENAMES.items():
        cps[fiscalsim_variable] = spm_unit[asec_variable]

    cps["reduced_price_school_meals_reported"] = (
        cps["free_school_meals_reported"][...] * 0
    )


def add_household_variables(cps: h5py.File, household: DataFrame) -> None:
    cps["state_fips"] = household.GESTFIPS
    cps["county_fips"] = household.GTCO
    state_county_fips = cps["state_fips"][...] * 1e3 + cps["county_fips"][...]
    # Assign is_nyc here instead of as a variable formula so that it shows up
    # as toggleable in the webapp.
    # List county FIPS codes for each NYC county/borough.
    NYC_COUNTY_FIPS = [
        5,  # Bronx
        47,  # Kings (Brooklyn)
        61,  # New York (Manhattan)
        81,  # Queens
        85,  # Richmond (Staten Island)
    ]
    # Compute NYC by concatenating NY state FIPS with county FIPS.
    # For example, 36061 is Manhattan.
    NYS_FIPS = 36
    nyc_full_county_fips = [
        NYS_FIPS * 1e3 + county_fips for county_fips in NYC_COUNTY_FIPS
    ]
    cps["in_nyc"] = np.isin(state_county_fips, nyc_full_county_fips)


class CPS_2020(CPS):
    name = "cps_2020"
    label = "CPS 2020"
    raw_cps = RawCPS_2020
    file_path = STORAGE_FOLDER / "cps_2020.h5"
    time_period = 2020


class CPS_2021(CPS):
    name = "cps_2021"
    label = "CPS 2021"
    raw_cps = RawCPS_2021
    file_path = STORAGE_FOLDER / "cps_2021.h5"
    time_period = 2021


class CPS_2022(CPS):
    name = "cps_2022"
    label = "CPS 2022"
    raw_cps = RawCPS_2022
    file_path = STORAGE_FOLDER / "cps_2022.h5"
    time_period = 2022
    # url = None


CPS_2023 = UpratedCPS.from_dataset(
    CPS_2022,
    2023,
    "cps_2023",
    "CPS 2023",
    STORAGE_FOLDER / "cps_2023.h5",
    new_url="release://TheCGO/fiscalsim-us/v0.2.9/cps_2023.h5",
)



from fiscalsim_us.model_api import *


class medicaid_benefit_value(Variable):
    value_type = float
    entity = Person
    label = "Average Medicaid payment"
    unit = USD
    documentation = "Per-capita payment for this person's State."
    definition_period = YEAR

    def formula(person, period, parameters):
        state = person.household("state_code_str", period)
        calibration = parameters(period).calibration
        spending = calibration.programs.medicaid.budgetary_impact[state]
        population = calibration.programs.medicaid.enrollment[state]
        return spending / population



from fiscalsim_us.model_api import *


class ut_recapture_low_inc_housing_credit(Variable):
    """
    Line 30 on Utah 2022 Individual Income Tax return form TC-40. This addition
    is described on form TC-40LIC
    (https://tax.utah.gov/forms/current/tc-40lic.pdf) and allows the filer to
    reapportion parts of the nonrefundable Utah low-income housing credit from
    a previous year filing for up to three years in order to capture the
    maximum amount possible of the credit.
    """

    value_type = float
    entity = TaxUnit
    label = "UT recapture of low-income housing credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



"""
This file defines the fiscalsim-us tax and benefit system.

A tax and benefit system is the higher-level instance in FiscalSim. Its goal is
to model the legislation of a country. Basically a tax and benefit system
contains simulation variables (source code) and legislation parameters (data).

See https://openfisca.org/doc/key-concepts/tax_and_benefit_system.html
"""

from fiscalsim_us.system import (
    CountryTaxBenefitSystem,
    Simulation,
    Microsimulation,
    IndividualSim,
)

from fiscalsim_us.data import *



from fiscalsim_us.model_api import *


class ar_metabolic_credit_carryover(Variable):
    """
    Line 2 on AR1113 PHENYLKETONURIA DISORDER AND OTHER METABOLIC DISORDERS CREDIT
    """

    value_type = float
    entity = TaxUnit
    label = "Arkansas metabolic credit carryover from previous year"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class tax_unit_dependent_elsewhere(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Is tax unit dependent elsewhere"
    documentation = "Whether the filer for this tax unit is claimed as a dependent in another tax unit."


dsi = variable_alias("dsi", tax_unit_dependent_elsewhere)



from fiscalsim_us.model_api import *


class tanf_amount_if_eligible(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF amount if family is eligible"
    documentation = "How much a family would receive if they were eligible for Temporary Assistance for Needy Families benefit."
    unit = USD

    def formula(spm_unit, period, parameters):
        max_amount = spm_unit("tanf_max_amount", period)
        countable_income = spm_unit("tanf_countable_income", period)
        return max_(0, max_amount - countable_income)



from fiscalsim_us.model_api import *


class ar_unreimbursed_employee_expenses(Variable):
    """
    Line 20 of 2022 AR3, Arkansas Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "AR unreimbursed employee expenses"
    unit = USD
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR2106_EmployeeBusinessExpenses.pdf"
    definition_period = YEAR
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class oh_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "OH taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf"
    defined_for = StateCode.OH



from fiscalsim_us.model_api import *


class ptc_phase_out_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "PTC phase-out rate"
    unit = "/1"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/36B#b_3_A"

    def formula(tax_unit, period, parameters):
        ptc = parameters(period).gov.irs.credits.premium_tax_credit
        income_level = tax_unit.value_from_first_person(
            tax_unit.members("medicaid_income_level", period)
        )
        rates = income_level * 0
        lower = ptc.phase_out.starting_rate
        upper = ptc.phase_out.ending_rate
        for i in range(len(lower.thresholds)):
            lower_threshold = lower.thresholds[i]
            upper_threshold = (
                lower.thresholds[i + 1]
                if i + 1 < len(lower.thresholds)
                else inf
            )
            in_bracket = (income_level >= lower_threshold) & (
                income_level < upper_threshold
            )
            percent_through_bracket = (income_level - lower_threshold) / (
                upper_threshold - lower_threshold
            )
            rate_in_bracket = (
                upper.amounts[i] - lower.amounts[i]
            ) * percent_through_bracket + lower.amounts[i]
            rates[in_bracket] = rate_in_bracket[in_bracket]
        return rates



from fiscalsim_us.model_api import *


class ut_taxpayer_credit_max(Variable):
    """
    Line 16 on Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT taxpayer credit maximum"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        gov = parameters(period).gov
        ut_taxpayer_credit = gov.states.ut.tax.income.credits.taxpayer_credit
        count_dependents = tax_unit("tax_unit_count_dependents", period)
        personal_exemption = (
            count_dependents * ut_taxpayer_credit.personal_exemption
        )

        itemizes = tax_unit("tax_unit_itemizes", period)
        us_standard_deduction = tax_unit("standard_deduction", period)
        deductions_if_itemizing = gov.irs.deductions.deductions_if_itemizing
        deductions_if_itemizing = [
            deduction
            for deduction in deductions_if_itemizing
            if deduction
            not in [
                "salt_deduction",
                # Exclude QBID to avoid circular reference.
                "qualified_business_income_deduction",
            ]
        ]
        us_itemized_deductions_less_salt = add(
            tax_unit, period, deductions_if_itemizing
        )

        total_deductions = personal_exemption + where(
            itemizes, us_itemized_deductions_less_salt, us_standard_deduction
        )

        return total_deductions * ut_taxpayer_credit.rate



from fiscalsim_us.model_api import *


class income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "Federal income tax before refundable credits"
    documentation = "Income tax liability (including other taxes) after non-refundable credits are used, but before refundable credits are applied"
    adds = [
        "income_tax_before_credits",
        "net_investment_income_tax",
        "recapture_of_investment_credit",
        "unreported_payroll_tax",
        "qualified_retirement_penalty",
    ]
    subtracts = ["income_tax_capped_non_refundable_credits"]



from fiscalsim_us.model_api import *


class ar_taxable_income(Variable):
    "Line 28 of Form AR1000F"
    value_type = float
    entity = TaxUnit
    label = "Arkansas taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/TaxBrackets_2022.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_Final_AR1000ES.pdf"
    )
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ar_agi", period)
        std_ded = tax_unit("ar_standard_deduction", period)
        itm_ded = tax_unit("ar_itemized_deductions", period)
        deduction = where(itm_ded > std_ded, itm_ded, std_ded)

        return agi - deduction



from fiscalsim_us.model_api import *


class ny_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (d)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        eitc = tax_unit("earned_income_tax_credit", period)
        rate = parameters(period).gov.states.ny.tax.income.credits.eitc.match
        tentative_nys_eic = eitc * rate
        household_credit = tax_unit("ny_household_credit", period)
        return max_(0, tentative_nys_eic - household_credit)



from fiscalsim_us.model_api import *


class mn_other_additions(Variable):
    """
    Unimplemented additions to Minnesota taxable income.
    Line 2 of Minnesota 2023 Individual Income Tax return form M1. These
    additions to income include the following categories which are not
    calculated with other variables.
    * Interest from municipal bond of another state
    * Federally exempt interest from mutual funds investing in state bonds
    * Expenses deducted on federal form income not taxed by Minnesota
    * Net nonqualified withdrawl from First-Time Homebuyer Savings Acct
    * Distributions from higher ed savings accout used in K-12 tuition
    """

    value_type = float
    entity = TaxUnit
    label = "MN additions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN
    # if implemented as separate vars, delete from list and add to sources yaml



from fiscalsim_us.model_api import *


class ctc_child_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "CTC maximum amount (child)"
    unit = USD
    documentation = "The CTC entitlement in respect of this person as a child."
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
    )

    def formula(person, period, parameters):
        age = person("age", period)
        base_amount = parameters(period).gov.irs.credits.ctc.amount.base
        is_dependent = person("is_tax_unit_dependent", period)
        return base_amount.calc(age) * is_dependent



from fiscalsim_us.model_api import *


class mt_additions_to_income(Variable):
    """
    Line 12 on Montana 2022 Individual Income Tax return form 2. These
    additions to income include the categories that are listed
    in Montana Additions Schedule on page 4 of Montana 2022 Individual Income Tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "MT additions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.data import CPS_2021, CPS_2022, CPS_2023
import pytest
from fiscalsim_us import Microsimulation
import warnings

warnings.filterwarnings("ignore")
warnings.simplefilter("ignore")

CPS_YEARS = []


@pytest.mark.dependency(name="cps")
@pytest.mark.parametrize("year", CPS_YEARS)
def test_cps_dataset_generates(year):
    year()


@pytest.mark.dependency(depends=["cps"])
@pytest.mark.parametrize("year", CPS_YEARS)
def test_cps_fiscalsim_us_compatible(year):
    Microsimulation(dataset=year).calc("employment_income")



from fiscalsim_us.model_api import *


class household_market_income(Variable):
    value_type = float
    entity = Household
    label = "market income"
    documentation = "Income from non-government sources."
    definition_period = YEAR
    unit = USD
    adds = [
        "employment_income",
        "self_employment_income",
        "pension_income",
        "dividend_income",
        "interest_income",
        "gi_cash_assistance",
        "capital_gains",
        "rental_income",
        "illicit_income",
        "farm_income",
        "miscellaneous_income",
    ]



from fiscalsim_us.model_api import *


class non_mn_bond_interest(Variable):
    """
    State and municipal bond interest from outside Minnesota
    not included in fed Form 6251
    Line 5 Minnesota 2022 form M1MT, Alternative Minimum Tax
    """

    value_type = float
    entity = TaxUnit
    label = "Non MN state and municipal bond interest"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class ia_modified_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa modified income used in tax-exempt and alternate-tax calculations"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=55"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=55"
    )
    defined_for = StateCode.IA
    adds = "gov.states.ia.tax.income.modified_income.sources"



from fiscalsim_us.model_api import *


class mn_depletion(Variable):
    """
    Line 7 Minnesota 2022 form M1MT, Alternative Minimum Tax
    """

    value_type = float
    entity = TaxUnit
    label = "MN depletion"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class pell_grant_head_contribution(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant head contribution"
    definition_period = YEAR

    def formula(person, period, parameters):
        available_income = person("pell_grant_head_available_income", period)
        dependents = person.tax_unit(
            "pell_grant_dependents_in_college", period
        )
        formula = person("pell_grant_formula", period).decode_to_str()
        p = parameters(period).gov.ed.pell_grant.efc.head
        positive_head_contribution = p.marginal_rate.calc(available_income)
        negative_head_contribution = max_(
            available_income * p.negative_rate, p.min_contribution
        )
        total_head_contribution = where(
            available_income >= 0,
            positive_head_contribution,
            negative_head_contribution,
        )
        total = where(
            formula == "B", available_income, total_head_contribution
        )
        # Return amount per dependent, using a mask to avoid division by zero.
        amount_per_dependent = np.zeros_like(total)
        mask = dependents > 0
        amount_per_dependent[mask] = total[mask] / dependents[mask]
        return amount_per_dependent



from fiscalsim_us.model_api import *


class ira_contributions(Variable):
    value_type = float
    entity = Person
    label = "IRA contributions"
    unit = USD
    documentation = "Contributions to Individual Retirement Accounts."
    definition_period = YEAR



from fiscalsim_us.model_api import *


class basic_income_before_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic income before phase-outs"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.contrib.ubi_center.basic_income.amount
        # Start with flat person-level amount.
        total_flat_amount = p.person.flat * tax_unit("tax_unit_size", period)
        # Add per-age person-level amount.
        person = tax_unit.members
        age = person("age", period)
        amount_by_age = p.person.by_age.calc(age)
        total_amount_by_age = tax_unit.sum(amount_by_age)
        # Now compute FPG amount.
        fpg = tax_unit("tax_unit_fpg", period)
        fpg_amount = p.tax_unit.fpg_percent * fpg
        return total_flat_amount + total_amount_by_age + fpg_amount



from fiscalsim_us.model_api import *


class ar_income_adjustments(Variable):
    """Line 24 of Form AR1000F
    Adjustments include:
        Border city/Texarkana exemption
        Tuition Savings Program
        Payments to IRA
        Payments to MSA
        Payments to HSA
        Deduction for interest paid on student loans
        Contributions to Intergenerational Trust
        Moving expenses
        Self-employed health insurance deduction
        KEOGH, Self-employed SEP and Simple Plans
        Forfeited interest penalty for premature withdrawal
        Alimony/Separate Maintenance Paid
        Support for permanently disabled individual
        Organ Donor Deduction
        Military Reserve Expenses
        Reforestation Deduction
        Teachers Qualified Classroom Investment Expense
        Achieving A Better Life Experience Program"""

    value_type = float
    entity = TaxUnit
    label = "Arkansas adjustments to total income to calculate AR AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000ADJ_AdjustmentsSchedule.pdf"
        "https://law.justia.com/codes/arkansas/2019/title-26/subtitle-5/chapter-51/subchapter-4/section-26-51-403/"
    )
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class cmbtp(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "AMT income not included in AGI"
    documentation = "Estimate of income considered for AMT but not AGI"
    unit = USD


class filer_cmbtp(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income on Form 6251 not in AGI for the tax unit (excluding dependents)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("cmbtp", tax_unit, period)


class e00200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Wages, salaries, and tips net of pension contributions"
    unit = USD

    def formula(person, period, parameters):
        return person("employment_income", period)


class filer_e00200(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Wages, salaries, and tips for filing unit (excluding dependents) net of pension contributions (pencon)"
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00200", tax_unit, period)


class pencon(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Contributions to defined-contribution pension plans"
    unit = USD


class filer_pencon(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Contributions to defined-contribution pension plans for filing unit (excluding dependents)"
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("pencon", tax_unit, period)


class e00300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Taxable interest income"
    unit = USD

    def formula(person, period, parameters):
        return person("interest_income", period)


class filer_e00300(Variable):
    value_type = float
    entity = TaxUnit
    label = "Taxable interest income for the tax unit (excluding dependents)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00300", tax_unit, period)


class e02300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Unemployment compensation benefits"
    unit = USD


class filer_e02300(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Unemployment compensation benefits for filing unit (excluding dependents)"
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum(
            "unemployment_compensation", tax_unit, period
        )


class e02400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Total Social Security (OASDI) benefits"
    unit = USD

    def formula(person, period, parameters):
        return person("social_security", period)


class tax_unit_ss(Variable):
    value_type = float
    entity = TaxUnit
    label = "Social security benefits for the tax unit (excluding dependents)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return add(tax_unit, period, ["social_security"])


class e87530(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Qualified tuition expenses"
    documentation = (
        "Adjusted qualified lifetime learning expenses for all students"
    )
    unit = USD


class elderly_dependents(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of dependents age 65+ in filing unit excluding taxpayer and spouse"


class incapable_of_self_care(Variable):
    value_type = bool
    entity = Person
    label = "Incapable of self-care"
    documentation = "Whether this person is physically or mentally incapable of caring for themselves."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21"


class cdcc_qualified_dependent(Variable):
    value_type = bool
    entity = Person
    label = "Qualifying dependent for CDCC"
    documentation = "Whether this person qualifies as a dependent for the child and dependent care credit."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21"

    def formula(person, period, parameters):
        cdcc = parameters(period).gov.irs.credits.cdcc
        meets_age_criteria = person("age", period) < cdcc.eligibility.child_age
        incapable_of_self_care = person("incapable_of_self_care", period)
        is_dependent = person("is_tax_unit_dependent", period)
        is_spouse = person("is_tax_unit_spouse", period)
        dependent_or_spouse = is_dependent | is_spouse
        return meets_age_criteria | (
            dependent_or_spouse & incapable_of_self_care
        )


class f2441(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of child/dependent-care qualifying persons"

    def formula(tax_unit, period, parameters):
        gross_num_eligible = tax_unit.sum(
            tax_unit.members("cdcc_qualified_dependent", period)
        )
        cdcc = parameters(period).gov.irs.credits.cdcc
        return min_(gross_num_eligible, cdcc.eligibility.max)


class f6251(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "True if Form 6251 (AMT) attached to return; otherwise false"
    )


class a_lineno(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "CPS line number for the person record of the head of the tax filing unit (not used in tax-calculation logic)"


class ffpos(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "CPS family identifier within household (not used in tax-calculation logic)"


class fips(Variable):
    value_type = int
    entity = Household
    definition_period = YEAR
    documentation = "FIPS state code (not used in tax-calculation logic)"
    default_value = 6


class h_seq(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "CPS household sequence number (not used in tax-calculation logic)"
    )


class data_source(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = "True if unit is created primarily from IRS-SOI PUF data; false if created primarily from CPS data (not used in tax-calculation logic)"


class k1bx14(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        "Partner self-employment earnings/loss (included in e26270 total)"
    )
    unit = USD


class filer_k1bx14(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Partner self-employment earnings/loss for tax unit (excluding dependents) (included in e26270 total)"
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("k1bx14", tax_unit, period)


class n24(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of children who are Child-Tax-Credit eligible, one condition for which is being under age 17"


class nu06(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of dependents under 6 years old"


class nu13(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of dependents under 13 years old"


class nu18(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of people under 18 years old in the filing unit"


class n1820(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of people age 18-20 years old in the filing unit"


class n21(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Number of people 21 years old or older in the filing unit"


class hasqdivltcg(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Has qualified dividends or long-term capital gains"
    documentation = "Whether this tax unit has qualified dividend income, or long-term capital gains income"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # Negatives cannot offset other income sources
        INCOME_SOURCES = [
            "c01000",
            "c23650",
            "long_term_capital_gains",
            "non_sch_d_capital_gains",
            "qualified_dividend_income",
        ]
        return np.any(
            [
                add(tax_unit, period, [income_source]) > 0
                for income_source in INCOME_SOURCES
            ]
        )


class c23650(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Net capital gains"
    unit = USD
    documentation = "Net capital gains (long and short term) before exclusion"

    def formula(tax_unit, period, parameters):
        return add(
            tax_unit,
            period,
            ["long_term_capital_gains", "short_term_capital_gains"],
        )



from fiscalsim_us.model_api import *


class oh_senior_citizen_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio senior citizen credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=20",
        "https://codes.ohio.gov/ohio-revised-code/section-5747.055",
    )
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.oh.tax.income.credits.senior_citizen
        person = tax_unit.members

        has_not_taken_lump_sum_distribution = person(
            "oh_has_not_taken_oh_lump_sum_credits", period
        )
        age = person("age", period)

        any_elderly = tax_unit.any(age >= p.age_threshold)
        credit_amount = p.agi_limit.calc(tax_unit("oh_agi", period))

        return (
            any_elderly * has_not_taken_lump_sum_distribution * credit_amount
        )



from fiscalsim_us.model_api import *


class housing_assistance(Variable):
    value_type = float
    entity = SPMUnit
    label = "Housing assistance"
    unit = USD
    documentation = "Housing assistance"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        hap = spm_unit("hud_hap", period)
        eligible = spm_unit("is_eligible_for_housing_assistance", period)
        return hap * eligible



from fiscalsim_us.model_api import *


label = "Income"


class employment_income(Variable):
    value_type = float
    entity = Person
    label = "employment income"
    documentation = "Wages and salaries, including tips and commissions."
    unit = USD
    definition_period = YEAR


class self_employment_income(Variable):
    value_type = float
    entity = Person
    label = "self-employment income"
    unit = USD
    documentation = "Self-employment non-farm income."
    definition_period = YEAR


class farm_income(Variable):
    value_type = float
    entity = Person
    label = "farm income"
    unit = USD
    documentation = "Income from agricultural businesses. Do not include this income in self-employment income."
    definition_period = YEAR



from fiscalsim_us.model_api import *


class spm_unit_payroll_tax_reported(Variable):
    value_type = float
    entity = SPMUnit
    label = "SPM unit payroll tax (reported)"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class ar_nonrefundable_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas Child and Dependent Care Credit Nonrefundable Portion"
    unit = USD
    documentation = "https://codes.findlaw.com/ar/title-26-taxation/ar-code-sect-26-51-502/"
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        cdcc = tax_unit("ar_cdcc", period)
        refundable_pct = tax_unit("ar_cdcc_refundable_pct", period)

        return cdcc * (1 - refundable_pct)



from fiscalsim_us.model_api import *


class mt_political_contributions_deduction(Variable):
    """
    Line 16 on itemized deductions schedule
    Limited to $100 per taxpayer
    """

    value_type = float
    entity = TaxUnit
    label = "Montana political contributions deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.deductions.itemized.misc
        cap = p.political_contribution_cap
        contributions = tax_unit("mt_political_contributions")

        return min_(cap, contributions)



from fiscalsim_us.model_api import *


class tanf_gross_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF gross earned income"
    documentation = "Gross earned income for calculating Temporary Assistance for Needy Families benefit."
    unit = USD
    reference = "https://www.dhs.state.il.us/page.aspx?item=15814"
    adds = "gov.hhs.tanf.cash.income.gross.sources.earned"



from fiscalsim_us.model_api import *


class va_tax_credit_for_low_income_individuals(Variable):
    value_type = float
    entity = TaxUnit
    label = "va tax credit for low income indivduals https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-adj-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("va_adj_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        fed_eitc = tax_unit("earned_income_tax_credit", period)
        net_tax = tax_unit("va_income_tax_before_refundable_credits", period)

        spouse_if_filing_jointly = where(
            (filing_status == filing_statuses.SEPARATE)
            | (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD),
            0,
            where(filing_status == filing_statuses.JOINT, 1, 0),
        )

        spouse_agi = tax_unit("spouse_separate_adjusted_gross_income", period)
        dependents = tax_unit("tax_unit_dependents", period)
        total_num_exemptions = spouse_if_filing_jointly + dependents + 1
        total_agi = agi + spouse_agi
        eitc_rate = parameters(period).gov.states.va.tax.income.va_eitc_rate
        threshold = parameters(
            period
        ).gov.states.va.tax.income.va_eitc_threshold.calc(total_num_exemptions)

        line_13 = where(total_agi < threshold, total_num_exemptions * 300, 0)
        line_14 = fed_eitc
        line_15 = line_14 * eitc_rate
        line_16 = where(line_15 > line_13, line_15, line_13)
        line_17 = where(net_tax > line_16, line_16, net_tax)

        return where(total_agi < threshold, line_17, 0)



from fiscalsim_us.model_api import *
import numpy as np


class ar_personal_credits(Variable):
    "Line 7A - 7D, 34 of form AR1000F"
    value_type = float
    entity = TaxUnit
    label = "Arkansas personal tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.credits.personal
        personal_credit_amount = p.personal_credit_amount
        person = tax_unit.members
        aged = person("age", period) >= p.age_threshold
        receives_retirement_or_disability_exemption = (
            person(
                "ar_retirement_or_disability_benefits_exemption_person", period
            )
            > 0
        )
        aged_special = aged & ~receives_retirement_or_disability_exemption
        # Only head and spouse are eligible for the personal credit amounts
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        # Blind filers get an additional personal tax credit amount
        blind = person("is_blind", period)
        # Deaf filers get an additional personal tax credit amount
        deaf = person("is_deaf", period)

        # Widowed and head of household filers receive an additional credit
        # amount
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        widow = filing_status == statuses.SURVIVING_SPOUSE
        hoh = filing_status == statuses.HEAD_OF_HOUSEHOLD
        whoh_filing_status_eligible = widow | hoh

        personal_credit_count = tax_unit.sum(
            head_or_spouse
            * (
                1
                + aged.astype(int)
                + blind.astype(int)
                + deaf.astype(int)
                + aged_special.astype(int)
            )
        ) + whoh_filing_status_eligible.astype(int)

        personal_credit = personal_credit_count * personal_credit_amount

        dependents = tax_unit("tax_unit_dependents", period)
        dependent_credit_amount = parameters(
            period
        ).gov.states.ar.tax.income.credits.personal.dependents
        dependent_credit = dependents * dependent_credit_amount

        qual_dependents = tax_unit(
            "ar_qual_dependents", period
        )  # Probably not implemented correctly
        qual_dependent_credit_amount = parameters(
            period
        ).gov.states.ar.tax.income.credits.personal.qual_dependents
        qual_dependent_credit = qual_dependent_credit_amount * qual_dependents

        return personal_credit + dependent_credit + qual_dependent_credit



import numpy as np
from fiscalsim_us.model_api import *


class is_ptc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Premium Tax Credit eligible"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        eligibility = parameters(
            period
        ).gov.irs.credits.premium_tax_credit.eligibility
        income_level = tax_unit("tax_unit_medicaid_income_level", period)
        on_marketplace = (
            add(tax_unit, period, ["has_marketplace_health_coverage"]) > 0
        )
        income_eligible = eligibility.calc(income_level)

        on_marketplace = np.array(on_marketplace, dtype=bool)
        income_eligible = np.array(income_eligible, dtype=bool)

        return on_marketplace & income_eligible



from fiscalsim_us.model_api import *


class mt_dependent_exemption_count(Variable):
    value_type = float
    entity = TaxUnit
    label = "Number of Montana dependent exemptions"
    unit = USD
    definition_period = YEAR
    reference = "https://regulations.justia.com/states/montana/department-42/chapter-42-15/subchapter-42-15-4/rule-42-15-403/"
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # Qualifying child under IRC 152(c), which defines for the EITC
        qualifying_child = person("is_eitc_qualifying_child", period)
        total_eligible = tax_unit.sum(qualifying_child)
        # Disabled dependents get an additional exemption.
        disabled = where(
            qualifying_child, person("is_disabled", period).astype(int), 0
        )
        total_disabled = tax_unit.sum(disabled)
        return total_eligible + total_disabled



from fiscalsim_us.model_api import *


class mn_other_subtractions(Variable):
    """
    Line 7 of Minnesota 2023 Individual Income Tax return from M1. These
    subtractions from income include the following categories which are
    not included as other variables directly.
    * Education expense paid for qulaifying children in grades K-12
    * Qualifying contributions to education savings plan
    * Railroad retirement benefits
    * Reservation income for American Indians
    * Federal activate duty military pay while Minnesota resident
    * Organ donor subtraction
    * Volunteer mileage reimbursment subtraction
    * Subtraction for miliary pensions and retirement
    * Payment from Minnesota Frontline Worker Pay Program
    """

    value_type = float
    entity = TaxUnit
    label = "MN other subtrations from income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    # if any get implimented as vars, delete from list



from fiscalsim_us.model_api import *


class va_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "va net amount of tax line 18 form 760 https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/760-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        rates = parameters(period).gov.states.va.tax.income.va_tax_rates
        taxable_inc = tax_unit("va_taxable_income", period)

        net_tax = where(taxable_inc > 0, rates.calc(taxable_inc), 0)

        return net_tax



from fiscalsim_us.model_api import *


class ar_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR3_ItemizedDeduction.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        medical_expense = add(tax_unit, period, ["medical_expense"])
        agi = tax_unit("ar_agi", period)
        medical_threshold = parameters(
            period
        ).gov.states.ar.tax.income.deductions.itemized.medical

        agi_portion = max_(0, agi * medical_threshold)

        medical_deduction = max_(0, medical_expense - agi_portion)

        real_estate_tax = add(tax_unit, period, ["real_estate_taxes"])

        # Line 6
        other_taxes = tax_unit("ar_other_taxes", period)

        interest_expense = tax_unit("interest_deduction", period)

        charitable_deduction = tax_unit("charitable_deduction", period)

        casualty_loss = add(tax_unit, period, ["casualty_loss"])

        education_deduction = tax_unit(
            "ar_post_secondary_education_tuition_deduction", period
        )

        # Miscellaneous deductions subject to AGI limit
        employee_expenses = tax_unit(
            "ar_unreimbursed_employee_expenses_deduction", period
        )
        other_limited_expenses = tax_unit("ar_other_limited_expenses", period)

        line_22 = employee_expenses + other_limited_expenses

        misc_limit_pct = parameters(
            period
        ).gov.states.ar.tax.income.deductions.itemized.miscellaneous_limit

        misc_limit = agi * misc_limit_pct

        misc_deductions = max_(0, line_22 - misc_limit)

        other_misc_deductions = tax_unit("ar_other_misc_deductions", period)

        return (
            medical_deduction
            + real_estate_tax
            + other_taxes
            + interest_expense
            + education_deduction
            + charitable_deduction
            + casualty_loss
            + misc_deductions
            + other_misc_deductions
        )



from fiscalsim_us.model_api import *


class ar_head_retirement_income(Variable):
    """If zero, mark '65 Special' Line 7A of form AR1000F
    - taxable_pension_income
    """

    value_type = float
    entity = TaxUnit
    label = "Arkansas retirement_income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_head = ~person("is_tax_unit_head", period)

        income = 0
        income += is_head * add(person, period, ["ar_retirement_sources"])

        return tax_unit.sum(income)



from fiscalsim_us.model_api import *


class state_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "state income tax"
    unit = USD
    definition_period = YEAR
    adds = [
        # state income tax variables listed in alphabetical order:
        "ar_income_tax",
        "ca_income_tax",
        "co_income_tax",
        "dc_income_tax",
        "ia_income_tax",
        "il_income_tax",
        "in_income_tax",
        "ks_income_tax",
        "ky_income_tax",
        "la_income_tax",
        "ma_income_tax",
        "md_income_tax",
        "me_income_tax",
        "mn_income_tax",
        "mt_income_tax",
        "mo_income_tax",
        "nc_income_tax",
        "nd_income_tax",
        "ne_income_tax",
        "nh_income_tax",
        "nj_income_tax",
        "nm_income_tax",
        "ny_income_tax",
        "ok_income_tax",
        "or_income_tax",
        "pa_income_tax",
        "ri_income_tax",
        "sc_income_tax",
        "ut_income_tax",
        "va_income_tax",
        "vt_income_tax",
        "wa_income_tax",
        "wi_income_tax",
    ]

    def formula(tax_unit, period, parameters):
        if parameters(period).simulation.reported_state_income_tax:
            spm_unit = tax_unit.spm_unit
            person = spm_unit.members
            is_head = person("is_tax_unit_head", period)
            total_tax_unit_heads = spm_unit.sum(is_head)
            spm_unit_state_tax = spm_unit(
                "spm_unit_state_tax_reported", period
            )
            return where(
                total_tax_unit_heads > 0,
                spm_unit_state_tax / total_tax_unit_heads,
                0,
            )
        else:
            return add(tax_unit, period, state_income_tax.adds)



from fiscalsim_us.model_api import *


class ar_qual_dependents(Variable):
    "Line 7C of form AR1000F"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Qualifying Dependents"
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_dependent = ~person("is_tax_unit_dependent", period)
        is_disabled = ~person("is_disabled", period)

        eligible = is_dependent * is_disabled

        return tax_unit.sum(eligible)



from fiscalsim_us.model_api import *


class az_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Itemized Deductions"
    unit = USD
    documentation = "Arizona Form 140 Schedule A"
    definition_period = YEAR
    defined_for = StateCode.AZ
    reference = (
        "https://law.justia.com/codes/arizona/2022/title-43/section-43-1042/",
        "https://azdor.gov/forms/individual/itemized-deduction-adjustments-form",
        "https://azdor.gov/forms/individual/form-140-resident-personal-income-tax-form-calculating",
    )
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        deductions = [
            deduction
            for deduction in p.itemized_deductions
            if deduction
            not in [
                "salt_deduction",
                "medical_expense_deduction",
                "charitable_deduction",
            ]
        ]
        federal_deductions = add(tax_unit, period, deductions)
        # Arizona allows a deduction for medical and dental expenses
        # that cannot be deducted from federal income taxes.
        medical_expense = add(tax_unit, period, ["medical_expense"])
        medical_expense_deduction = tax_unit(
            "medical_expense_deduction", period
        )

        az_medical_expense_deduction = max_(
            0, medical_expense - medical_expense_deduction
        )

        # Adjustments to Charitable Contributions
        # Amount of charitable contributions for which you are claiming
        # a credit under Arizona law.
        charitable_deduction = tax_unit("charitable_deduction", period)
        charitable_contributions_credit = tax_unit(
            "az_charitable_contributions_credit", period
        )

        charitable_deduction_after_credit = max_(
            charitable_deduction - charitable_contributions_credit, 0
        )

        # Adjustment to State Income Taxes - add back real estate taxes
        real_estate_taxes = add(tax_unit, period, ["real_estate_taxes"])

        return (
            federal_deductions
            + az_medical_expense_deduction
            + charitable_deduction_after_credit
            + real_estate_taxes
        )



from fiscalsim_us.model_api import *


class ut_taxpayer_credit(Variable):
    """
    Line 20 of Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT taxpayer credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    adds = ["ut_taxpayer_credit_max"]
    subtracts = ["ut_taxpayer_credit_reduction"]



from fiscalsim_us.model_api import *


class snap_child_support_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP child support payment deduction"
    unit = USD
    documentation = (
        "Deduction from SNAP gross income for child support payments"
    )
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_4"

    adds = ["child_support_expense"]



from fiscalsim_us.model_api import *


class is_widowed(Variable):
    value_type = bool
    entity = Person
    label = "Widowed"
    documentation = "Whether the person is widowed."
    definition_period = YEAR



from fiscalsim_us.model_api import *


class salt_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "SALT deduction"
    unit = USD
    documentation = "State and local taxes paid plus real estate tax deduction from taxable income."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/164"

    def formula(tax_unit, period, parameters):
        salt_amount = add(
            tax_unit,
            period,
            [
                "statelocal_sales_or_prior_inctax",
                "real_estate_taxes",
            ],
        )
        salt = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate
        cap = salt.cap[tax_unit("filing_status", period)]
        return min_(cap, salt_amount)



from fiscalsim_us.model_api import *


class nj_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=45"
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nj.tax.income.credits.ctc

        # Get amount per qualifying child based on taxable income.
        taxable_income = tax_unit("nj_taxable_income", period)
        amount_per_qualifying_child = p.amount.calc(taxable_income)

        # Get number of eligible children dependents
        person = tax_unit.members
        meets_age_limit = person("age", period) < p.ineligible_age
        dependent = person("is_tax_unit_dependent", period)
        age_dependent_eligible = meets_age_limit & dependent
        count_eligible = tax_unit.sum(age_dependent_eligible)

        # Exclude married filing separately filers.
        filing_status = tax_unit("filing_status", period)
        filing_eligible = (
            filing_status != filing_status.possible_values.SEPARATE
        )

        # Calculate total child tax credit
        return count_eligible * amount_per_qualifying_child * filing_eligible



from fiscalsim_us.model_api import *


class interest_expense(Variable):
    value_type = float
    entity = Person
    label = "Interest paid on all loans"
    unit = USD
    definition_period = YEAR
    adds = ["mortgage_interest", "non_mortgage_interest"]



from fiscalsim_us.model_api import *


class would_claim_wic(Variable):
    value_type = bool
    entity = Person
    label = "Would claim WIC"
    definition_period = YEAR

    def formula(person, period, parameters):
        if person.count < 1_000:
            # Don't run takeup imputations if not in a microsimulation.
            return True
        category = person("wic_category", period)
        takeup = parameters(period).gov.usda.wic.takeup
        return random(person) < takeup[category]



from fiscalsim_us.model_api import *


class maximum_state_supplement(Variable):
    value_type = float
    entity = Person
    label = "Maximum State Supplement"
    unit = USD
    definition_period = YEAR
    defined_for = "is_ssi_eligible_individual"

    def formula_2022(person, period, parameters):
        marital_unit = person.marital_unit
        eligible = person("is_ssi_eligible_individual", period)
        state_code = person.household("state_code_str", period)
        living_arrangement = person.household(
            "state_living_arrangement", period
        )
        ss_amounts = parameters(period).gov.ssa.state_supplement.amount
        amounts = ss_amounts[state_code][living_arrangement]
        is_blind = person("is_blind", period)
        is_aged = person("is_ssi_aged", period)
        is_disabled = person("is_ssi_disabled", period)
        count_persons = len(is_blind)
        ssi_categories = person("ssi_category", period).possible_values
        joint_claim = person("ssi_claim_is_joint", period)
        count_eligible_str = where(joint_claim, 2, 1).astype(str)
        per_person_amount = (
            max_(
                amounts[np.array([ssi_categories.AGED] * count_persons)][
                    count_eligible_str
                ]
                * is_aged,
                amounts[np.array([ssi_categories.BLIND] * count_persons)][
                    count_eligible_str
                ]
                * is_blind,
                amounts[np.array([ssi_categories.DISABLED] * count_persons)][
                    count_eligible_str
                ]
                * is_disabled,
            )
            * MONTHS_IN_YEAR
        )
        combined_amount = marital_unit.sum(per_person_amount)
        return eligible * combined_amount * where(joint_claim, 1 / 2, 1)



from fiscalsim_us.model_api import *


class spm_unit_medical_expenses(Variable):
    value_type = float
    entity = SPMUnit
    label = "SPM unit medical expenses"
    definition_period = YEAR
    unit = USD

    adds = ["medical_expense"]



from fiscalsim_us.model_api import *


class va_credits_from_enclosed_schedule_cr(Variable):
    value_type = float
    entity = TaxUnit
    label = "Schedule CR, Section 5, Part 1, Line 1A"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA



from fiscalsim_us.model_api import *


class dc_tanf_countable_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF countable earned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        gross_earnings = spm_unit("dc_tanf_gross_earned_income", period)
        p = parameters(period).gov.states.dc.dhs.tanf.income.deductions.earned
        enrolled = spm_unit("is_tanf_enrolled", period)
        annual_flat_exclusion = p.flat * MONTHS_IN_YEAR
        earnings_after_flat_exclusion = max_(
            gross_earnings - annual_flat_exclusion, 0
        )
        return where(
            enrolled,
            # For enrolled recipients, DC applies a flat and a percent deduction.
            earnings_after_flat_exclusion * (1 - p.percentage),
            # For new applicants, DC applies only a flat deduction.
            earnings_after_flat_exclusion,
        )



from fiscalsim_us.model_api import *


class sc_parentrefund_credit(Variable):
    """
    Parental refund credit, refundable credit, line 22d on SC1040 2023
    Invidual Income Tax Return (see form I-361)
    """

    value_type = float
    entity = Person
    label = "South Carolina Parent Refund Credit"
    defined_for = "sc_tuition_credit_eligible"
    unit = USD
    definition_period = YEAR
    reference = ("https://dor.sc.gov/forms-site/Forms/I361_2023.pdf",)



import torch
import pandas as pd
import numpy as np
from fiscalsim_us import Microsimulation
import plotly.express as px
from tqdm import tqdm
from policyengine_core.data import Dataset
from fiscalsim_us.data import CPS_2023
import numpy as np
from fiscalsim_us import Microsimulation
import pandas as pd
from .process_puf import (
    FINANCIAL_SUBSET as FINANCIAL_VARIABLES,
    puf_imputed_cps_person_level,
)
from typing import Tuple


def generate_model_variables(dataset: str, time_period: str = "2023") -> Tuple:
    """Generates variables needed for the calibration process.

    Args:
        dataset (str): The name of the dataset to use.
        time_period (str, optional): The time period to use. Defaults to "2023".

    Returns:
        household_weights (torch.Tensor): The household weights.
        weight_adjustment (torch.Tensor): The weight adjustment.
        values_df (pd.DataFrame): A 2D array of values to transform household weights into statistical predictions.
        targets (dict): A dictionary of names and target values for the statistical predictions.
        targets_array (dict): A 1D array of target values for the statistical predictions.
        equivalisation_factors_array (dict): A 1D array of equivalisation factors for the statistical predictions to normalise the targets.
    """
    simulation = Microsimulation(dataset=dataset)
    simulation.default_calculation_period = time_period
    parameters = simulation.tax_benefit_system.parameters.calibration(
        f"{time_period}-01-01"
    )

    household_weights = torch.tensor(
        simulation.calculate("household_weight").values, dtype=torch.float32
    )
    weight_adjustment = torch.tensor(
        np.random.random(household_weights.shape) * 10,
        requires_grad=True,
        dtype=torch.float32,
    )

    values_df = pd.DataFrame()
    targets = {}
    equivalisation = {}

    # We need to normalise the targets. Common regression targets are often 1e1 to 1e3 (this informs the scale of the learning rate).
    COUNT_HOUSEHOLDS = household_weights.sum().item()
    FINANCIAL_EQUIVALISATION = COUNT_HOUSEHOLDS
    POPULATION_EQUIVALISATION = COUNT_HOUSEHOLDS / 1e5

    for variable_name in FINANCIAL_VARIABLES:
        if variable_name not in parameters.gov.irs.soi:
            continue
        label = (
            simulation.tax_benefit_system.variables[variable_name].label
            + " aggregate"
        )
        values_df[label] = simulation.calculate(
            variable_name, map_to="household"
        ).values
        targets[label] = parameters.gov.irs.soi[variable_name]
        equivalisation[label] = FINANCIAL_EQUIVALISATION

    # Program spending from CBO baseline projections

    PROGRAMS = [
        "income_tax",
        "snap",
        "social_security",
        "ssi",
        "unemployment_compensation",
    ]

    for variable_name in PROGRAMS:
        label = simulation.tax_benefit_system.variables[variable_name].label
        values_df[label] = simulation.calculate(
            variable_name, map_to="household"
        ).values
        targets[label] = parameters.gov.cbo[variable_name]
        equivalisation[label] = FINANCIAL_EQUIVALISATION

    snap_participation = parameters.gov.usda.snap.participation
    ssi_participation = parameters.gov.ssa.ssi.participation
    ss_participation = parameters.gov.ssa.social_security.participation

    for program, participation in zip(
        ["snap", "ssi", "social_security"],
        [snap_participation, ssi_participation, ss_participation],
    ):
        label = simulation.tax_benefit_system.variables[program].label
        entity_level = simulation.tax_benefit_system.variables[
            program
        ].entity.key
        entity_level_value = simulation.calculate(program)
        values_df[f"{label} participants"] = simulation.map_result(
            entity_level_value > 0, entity_level, "household"
        )
        targets[f"{label} participants"] = participation
        equivalisation[f"{label} participants"] = POPULATION_EQUIVALISATION

    # Number of tax returns by AGI size

    agi_returns_thresholds = (
        parameters.gov.irs.soi.agi.number_of_returns.thresholds
    )
    agi_returns_values = parameters.gov.irs.soi.agi.number_of_returns.amounts
    agi = simulation.calculate("adjusted_gross_income").values
    is_filer = simulation.calculate("income_tax").values != 0
    for i in range(len(agi_returns_thresholds)):
        lower = agi_returns_thresholds[i]
        if i == len(agi_returns_thresholds) - 1:
            upper = np.inf
        else:
            upper = agi_returns_thresholds[i + 1]

        in_range = (agi >= lower) * (agi < upper) * is_filer
        household_returns_in_range = simulation.map_result(
            in_range, "tax_unit", "household"
        )

        name = f"Tax returns with ${lower:,.0f} <= AGI < ${upper:,.0f}"
        values_df[name] = household_returns_in_range
        targets[name] = agi_returns_values[i]
        equivalisation[name] = POPULATION_EQUIVALISATION
    # Total AGI aggregate by AGI band

    agi_returns_thresholds = parameters.gov.irs.soi.agi.total_agi.thresholds
    agi_returns_values = parameters.gov.irs.soi.agi.total_agi.amounts
    for i in range(len(agi_returns_thresholds)):
        lower = agi_returns_thresholds[i]
        if i == len(agi_returns_thresholds) - 1:
            upper = np.inf
        else:
            upper = agi_returns_thresholds[i + 1]

        in_range = (agi >= lower) * (agi < upper) * is_filer
        agi_in_range = agi * in_range
        household_agi_in_range = simulation.map_result(
            agi_in_range, "tax_unit", "household"
        )

        name = f"Total AGI from tax returns with ${lower:,.0f} <= AGI < ${upper:,.0f}"
        values_df[name] = household_agi_in_range
        targets[name] = agi_returns_values[i]
        equivalisation[name] = FINANCIAL_EQUIVALISATION

    # Total population
    values_df["U.S. population"] = simulation.calculate(
        "people", map_to="household"
    ).values
    targets["U.S. population"] = parameters.populations.total
    equivalisation["U.S. population"] = POPULATION_EQUIVALISATION

    # Population by 10-year age group and sex
    age = simulation.calculate("age").values
    is_male = simulation.calculate("is_male")
    for lower_age_group in range(0, 90, 10):
        for possible_is_male in (True, False):
            in_age_range = (age >= lower_age_group) & (
                age < lower_age_group + 5
            )
            in_sex_category = is_male == possible_is_male
            count_people_in_range = simulation.map_result(
                in_age_range * in_sex_category, "person", "household"
            )
            sex_category = "male" if possible_is_male else "female"
            name = f"{lower_age_group} to {lower_age_group + 5} and {sex_category} population"
            values_df[name] = count_people_in_range
            targets[name] = (
                household_weights.numpy() * count_people_in_range
            ).sum()
            equivalisation[name] = POPULATION_EQUIVALISATION

    # Household population by number of adults and children

    household_count_adults = simulation.map_result(
        age >= 18, "person", "household"
    )
    household_count_children = simulation.map_result(
        age < 18, "person", "household"
    )

    for count_adults in range(1, 3):
        for count_children in range(0, 4):
            in_criteria = (
                (household_count_adults == count_adults)
                * (household_count_children == count_children)
                * 1.0
            )
            name = f"{count_adults}-adult, {count_children}-child household population"
            values_df[name] = in_criteria
            targets[name] = (household_weights.numpy() * in_criteria).sum()
            equivalisation[name] = POPULATION_EQUIVALISATION

    # Tax filing unit counts by filing status

    filing_status = simulation.calculate("filing_status").values

    for filing_status_value in np.unique(filing_status):
        is_filing_status = filing_status == filing_status_value
        name = f"Filing status {filing_status_value.lower()} population"
        household_filing_status_unit_counts = simulation.map_result(
            is_filing_status, "tax_unit", "household"
        )
        values_df[name] = household_filing_status_unit_counts
        targets[name] = (
            household_weights.numpy() * household_filing_status_unit_counts
        ).sum()
        equivalisation[name] = POPULATION_EQUIVALISATION

    targets_array = torch.tensor(list(targets.values()), dtype=torch.float32)
    equivalisation_factors_array = torch.tensor(
        list(equivalisation.values()), dtype=torch.float32
    )

    return (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    )


def aggregate(
    adjusted_weights: torch.Tensor, values: pd.DataFrame
) -> torch.Tensor:
    broadcasted_weights = adjusted_weights.reshape(-1, 1)
    weighted_values = torch.matmul(
        broadcasted_weights.T, torch.tensor(values.values, dtype=torch.float32)
    )
    return weighted_values


def calibrate(
    dataset: str,
    time_period: str = "2023",
    training_log_path: str = "training_log.csv.gz",
    learning_rate: float = 1e-1,
    epochs: int = 250_000,
) -> np.ndarray:
    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables(dataset, time_period)
    training_log_df = pd.DataFrame()

    progress_bar = tqdm(range(epochs), desc="Calibrating weights")
    starting_loss = None
    for i in progress_bar:
        adjusted_weights = torch.relu(household_weights + weight_adjustment)
        result = (
            aggregate(adjusted_weights, values_df)
            / equivalisation_factors_array
        )
        loss = torch.mean(
            (result - targets_array / equivalisation_factors_array) ** 2
        )
        if i == 0:
            starting_loss = loss.item()
        loss.backward()
        if i % 20 == 0:
            current_loss = loss.item()
            progress_bar.set_description_str(
                f"Calibrating weights | Loss = {current_loss:,.0f}"
            )
            current_aggregates = (
                (result * equivalisation_factors_array).detach().numpy()[0]
            )
            training_log_df = pd.concat(
                [
                    training_log_df,
                    pd.DataFrame(
                        {
                            "name": list(targets.keys()) + ["total"],
                            "epoch": [i] * len(targets) + [i],
                            "value": list(current_aggregates) + [current_loss],
                            "target": list(targets.values()) + [0],
                        }
                    ),
                ]
            )
        weight_adjustment.data -= learning_rate * weight_adjustment.grad
        weight_adjustment.grad.zero_()

    training_log_df.to_csv(training_log_path, compression="gzip")

    loss_reduction = loss.item() / starting_loss - 1

    print(f"Loss reduction: {loss_reduction:.2%}")

    return adjusted_weights.detach().numpy()



from fiscalsim_us.model_api import *


class premium_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Premium Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/36B"

    def formula(tax_unit, period, parameters):
        applicable_percentage = tax_unit("ptc_phase_out_rate", period)
        eligible = tax_unit("is_ptc_eligible", period)
        plan_cost = tax_unit("second_lowest_silver_plan_cost", period)
        income = tax_unit("medicaid_income", period)
        return eligible * max_(0, plan_cost - income * applicable_percentage)



from fiscalsim_us.model_api import *


class ar_retirement_sources(Variable):
    value_type = float
    entity = Person
    label = "Income from Qualified IRA distributions and Employer pensions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    adds = "gov.states.ar.tax.income.retirement_sources"



from fiscalsim_us.model_api import *


class dc_tanf_countable_gross_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF countable gross unearned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.tanf.income
        # Sum unearned sources, plus child support disregard $150 received per month.
        gross_unearned = add(spm_unit, period, p.sources.unearned)
        child_support = add(spm_unit, period, ["child_support_received"])
        monthly_child_support_deduction = p.deductions.child_support
        child_support_after_deduction = max_(
            child_support - monthly_child_support_deduction * MONTHS_IN_YEAR, 0
        )
        return gross_unearned + child_support_after_deduction



from fiscalsim_us.model_api import *


class is_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Eligibility for TANF"
    documentation = "Whether the family is eligible for Temporary Assistance for Needy Families benefit."

    def formula(spm_unit, period, parameters):
        demographic_eligible = spm_unit("is_demographic_tanf_eligible", period)
        economic_eligible = where(
            spm_unit("is_tanf_enrolled", period),
            spm_unit("is_tanf_continuous_eligible", period),
            spm_unit("is_tanf_initial_eligible", period),
        )
        return demographic_eligible & economic_eligible



from fiscalsim_us.model_api import *


class ar_cdcc_refundable_pct(Variable):
    value_type = float
    entity = TaxUnit
    label = "Percentage of AR CCDC that is refundable, meaning that it quaifies for the Early Childhood Program"
    documentation = "https://codes.findlaw.com/ar/title-26-taxation/ar-code-sect-26-51-502/"
    definition_period = YEAR
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class is_ssi_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Indicates whether a person is disabled for the Supplemental Security Income program"
    label = "SSI disabled"
    reference = "https://www.law.cornell.edu/uscode/text/42/1382c#a_3_A"

    def formula(person, period, parameters):
        aged = person("is_ssi_aged", period)
        blind = person("is_blind", period)
        reported_receipt = person("ssi_reported", period) > 0
        return reported_receipt & ~aged & ~blind



from fiscalsim_us.model_api import *


class dc_tanf_countable_resources(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF countable resources"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC



from fiscalsim_us import Simulation, Microsimulation
from policyengine_core.tracers import FullTracer

simulation = Microsimulation()
simulation.trace = True
simulation.calculate("household_net_income", 2023)
simulation.tracer.generate_performance_graph(".")



from fiscalsim_us.model_api import *


class mn_disabled_impairment_work_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MN impairment-related work expenses for disabled person"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class in_nyc(Variable):
    value_type = bool
    entity = Household
    definition_period = YEAR
    label = "Is in NYC"
    # Currently disabled to allow entry until we collect county.
    # County of the below list.
    # NYC_COUNTIES = [
    #   "NEW_YORK_COUNTY_NY",
    #   "KINGS_COUNTY_NY",
    #   "QUEENS_COUNTY_NY",
    #   "RICHMOND_COUNTY_NY",
    #   "BRONX_COUNTY_NY",
    # ]



from fiscalsim_us.model_api import *


class earned_income_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "EITC"
    reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
    unit = USD
    defined_for = "eitc_eligible"

    def formula(tax_unit, period, parameters):
        maximum = tax_unit("eitc_maximum", period)
        phased_in = tax_unit("eitc_phased_in", period)
        reduction = tax_unit("eitc_reduction", period)
        limitation = max_(0, maximum - reduction)
        return min_(phased_in, limitation)


c59660 = variable_alias("c59660", earned_income_tax_credit)
eitc = variable_alias("eitc", earned_income_tax_credit)



from fiscalsim_us.model_api import *


class ny_supplemental_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY Supplemental EITC"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (d)(8)

    def formula(tax_unit, period, parameters):
        eitc = tax_unit("earned_income_tax_credit", period)
        rate = parameters(
            period
        ).gov.states.ny.tax.income.credits.eitc.supplemental_match
        return eitc * rate



from fiscalsim_us.model_api import *


class me_step_4_share_of_child_care_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine Step 4 Share of Child Care Expenses"
    documentation = "Share of child care expenses that are from programs classified as Step 4 in Maine"
    definition_period = YEAR
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html"



from fiscalsim_us.model_api import *


class mt_child_care_expense_deduction(Variable):
    """
    Line 14 on Montana itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana child and dependent care expenses"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class dc_tanf_grant_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF grant standard"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        unit_size = spm_unit("spm_unit_size", period)
        p = parameters(period).gov.states.dc.dhs.tanf.grant_standard
        monthly = p.main[unit_size]
        return monthly * MONTHS_IN_YEAR



from fiscalsim_us.model_api import *


class va_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "line 26 on the va tax return form 760"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        net_tax = tax_unit("va_income_tax_before_refundable_credits", period)
        line_26 = tax_unit("va_refundable_credits", period)

        owed_tax = where(line_26 < net_tax, net_tax - line_26, 0)
        refund = where(net_tax < line_26, (line_26 - net_tax) * -1, 0)

        return owed_tax + refund



from fiscalsim_us.model_api import *


class zip_code(Variable):
    value_type = str
    entity = Household
    label = "ZIP code"
    definition_period = YEAR
    default_value = "UNKNOWN"

    def formula(household, period, parameters):
        numeric_zip_code = ZIP_CODE_DATASET.zip_code.sample(
            household.count, weights=ZIP_CODE_DATASET.population, replace=True
        )
        return numeric_zip_code.astype(str).str.zfill(5)



from policyengine_core.data import Dataset
from fiscalsim_us.data.datasets.cps.cps import CPS_2023
from fiscalsim_us.data.storage import STORAGE_FOLDER
import numpy as np
import pandas as pd


class PUFExtendedCPS_2023(Dataset):
    name = "puf_extended_cps_2023"
    label = "PUF-extended CPS (2023)"
    file_path = STORAGE_FOLDER / "puf_extended_cps_2023.h5"
    data_format = Dataset.ARRAYS
    time_period = "2023"

    def generate(self):
        from .process_puf import (
            FINANCIAL_SUBSET as FINANCIAL_VARIABLES,
            puf_imputed_cps_person_level,
        )

        person_df = puf_imputed_cps_person_level()
        new_data = {}
        cps = CPS_2023()
        cps_data = cps.load()
        for variable in list(set(cps.variables) | set(FINANCIAL_VARIABLES)):
            if "_id" in variable:
                # Append on a copy multiplied by 10
                new_data[variable] = np.concatenate(
                    [cps_data[variable][...], cps_data[variable][...] + 1e8]
                )
            elif "_weight" in variable:
                # Append on a zero-weighted copy
                new_data[variable] = np.concatenate(
                    [
                        cps_data[variable][...],
                        np.zeros_like(cps_data[variable][...]),
                    ]
                )
            else:
                # Append on a copy
                if variable in FINANCIAL_VARIABLES:
                    if variable not in cps.variables:
                        if variable == "social_security":
                            ## SS is an edge case
                            original_values = (
                                cps_data["social_security_retirement"][...]
                                + cps_data["social_security_disability"][...]
                            )
                        else:
                            original_values = np.zeros_like(
                                cps_data["employment_income"][...]
                            )
                    else:
                        original_values = cps_data[variable][...]
                    new_data[variable] = np.concatenate(
                        [original_values, person_df[variable].values]
                    )
                else:
                    new_data[variable] = np.concatenate(
                        [cps_data[variable][...], cps_data[variable][...]]
                    )

        self.save_dataset(new_data)



from fiscalsim_us.model_api import *


class snap(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Final SNAP benefit amount, equal to net income minus food contribution"
    label = "SNAP allotment"
    reference = "https://www.law.cornell.edu/uscode/text/7/2017#a"
    unit = USD
    adds = ["snap_normal_allotment", "snap_emergency_allotment"]

    def formula(spm_unit, period, parameters):
        if parameters(period).gov.usda.snap.abolish_snap:
            return 0
        else:
            added_components = add(spm_unit, period, snap.adds)
            return added_components



from fiscalsim_us.model_api import *


class de_elderly_or_disabled_income_exclusion_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Delaware elderly or disabled income exclusion"
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        # First get their filing status.
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.de.tax.income.subtractions.exclusions.elderly_or_disabled

        # Get the individual disabled status.
        disabled_head = tax_unit("disabled_head", period)
        disabled_spouse = tax_unit("disabled_spouse", period)

        # Get the individual filer's age and eligibility.
        age_head = tax_unit("age_head", period)
        age_head_eligible = age_head > p.eligibility.age_threshold

        # Get spouse age and eligibility
        age_spouse = tax_unit("age_spouse", period)
        age_spouse_eligible = age_spouse > p.eligibility.age_threshold

        # Get the tax unit income
        earned_income = tax_unit("tax_unit_earned_income", period)

        # Determine if filer income is eligible.
        income_threshold = p.eligibility.earned_income_limit[filing_status]
        income_eligible = earned_income < income_threshold

        # Check the individual's eligibility.
        is_joint = tax_unit("tax_unit_is_joint", period)
        head_eligible = disabled_head | age_head_eligible
        spouse_eligible = disabled_spouse | age_spouse_eligible
        age_or_disability_eligible = where(
            is_joint,
            head_eligible & spouse_eligible,
            head_eligible,
        )

        pre_exclusions_agi = tax_unit("de_pre_exclusions_agi", period)
        agi_eligible = (
            pre_exclusions_agi <= p.eligibility.agi_limit[filing_status]
        )

        return age_or_disability_eligible & income_eligible & agi_eligible



from fiscalsim_us.model_api import *


class ia_regular_tax_indiv(Variable):
    value_type = float
    entity = Person
    label = "Iowa regular tax calculated using income tax rate schedule when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=53"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        taxable_income = person("ia_taxable_income_indiv", period)
        p = parameters(period).gov.states.ia.tax.income
        return p.rates.all.calc(max_(0, taxable_income))



from fiscalsim_us.model_api import *


class four_year_college_student(Variable):
    value_type = bool
    entity = Person
    label = "Is a four year university/college student"
    definition_period = YEAR



from fiscalsim_us.model_api import *


class tanf(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF"
    documentation = (
        "Value of Temporary Assistance for Needy Families benefit received."
    )
    unit = USD

    def formula(spm_unit, period, parameters):
        if parameters(period).gov.hhs.tanf.abolish_tanf:
            return 0
        tanf_reported = add(spm_unit, period, ["tanf_reported"])
        if tanf_reported.sum() > 0:
            return tanf_reported
        # First compute TANF for states with it defined in gov/hhs/tanf.
        # This is IL and CA.
        # (We will move these into gov/states)
        # Obtain eligibility.
        eligible = spm_unit("is_tanf_eligible", period)
        # Obtain amount they would receive if they were eligible.
        amount_if_eligible = spm_unit("tanf_amount_if_eligible", period)
        # Add TANF programs computed in variables/gov/states.
        STATES_WITH_TANF = ["co", "ny", "dc"]
        state_tanf = add(
            spm_unit, period, [i + "_tanf" for i in STATES_WITH_TANF]
        )
        return where(eligible, amount_if_eligible, 0) + state_tanf



from fiscalsim_us.model_api import *


class co_state_addback(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado state income tax addback"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2021.pdf#page=5",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=5",
        "https://tax.colorado.gov/sites/tax/files/documents/ITT_State_Income_Tax_Addback_Jan_2023.pdf",
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        federal_itemizer = tax_unit("tax_unit_itemizes", period)

        salt_deduct = tax_unit("salt_deduction", period)
        local_taxes = add(
            tax_unit, period, ["prior_year_local_income_tax_paid"]
        )
        property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        state_addback = max_(0, salt_deduct - local_taxes - property_taxes)

        # return the max between the "extra" itemized deductions above the std
        # and the state taxes included in the SALT deduction
        p = parameters(period).gov.irs.deductions
        item_deducts = add(tax_unit, period, p.itemized_deductions)
        std_deducts = tax_unit("standard_deduction", period)
        alt_addback = max_(0, item_deducts - std_deducts)

        return federal_itemizer * min_(state_addback, alt_addback)



from fiscalsim_us.model_api import *


class per_vehicle_payment(Variable):
    value_type = float
    entity = Person
    label = "Per-vehicle payment"
    unit = USD
    documentation = (
        "Payment to vehicle owners in respect of each owned vehicle."
    )
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        per_vehicle = parameters(period).gov.states.ca.per_vehicle_payment
        state = person.household("state_code", period)
        vehicles_owned = person("vehicles_owned", period)
        capped_vehicles = min_(vehicles_owned, per_vehicle.max_vehicles[state])
        return capped_vehicles * per_vehicle.amount[state]



import numpy as np
from fiscalsim_us import Microsimulation
from policyengine_core.reforms import Reform
from policyengine_core.periods import instant
import pytest


"""
In US nationwide simulations, use reported state income tax liabilities
"""


def use_reported_state_income_tax(parameters):
    parameters.simulation.reported_state_income_tax.update(
        start=instant("2024-01-01"), stop=instant("2100-12-31"), value=True
    )
    return parameters


class baseline_reform(Reform):
    def apply(self):
        self.modify_parameters(use_reported_state_income_tax)


@pytest.mark.skipif(True, reason="This test temporarily suspended.")
def test_microsim_runs_cps():
    sim = Microsimulation(reform=baseline_reform)
    hnet = sim.calc("household_net_income", period=2024, map_to="person")
    assert not hnet.isna().any(), "Some households have NaN net income."
    hidecile = sim.calc(
        "household_income_decile", period=2024, map_to="person"
    )
    assert np.all(hidecile >= 1) and np.all(hidecile <= 10)
    sidecile = sim.calc("spm_unit_income_decile", period=2024, map_to="person")
    assert np.all(sidecile >= 1) and np.all(sidecile <= 10)
    idecile = sim.calc("income_decile", period=2024, map_to="person")
    assert np.all(idecile >= 1) and np.all(idecile <= 10)



from fiscalsim_us.model_api import *


class mt_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana exemptions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        exemption_count = tax_unit("mt_exemptions_count", period)
        p = parameters(period).gov.states.mt.tax.income.exemptions
        return exemption_count * p.amount



from fiscalsim_us.model_api import *


class ar_other_net_gains(Variable):
    value_type = float
    entity = Person
    label = "Other net gains"
    unit = USD
    documentation = "Other net gain/loss from Federal Form 4797"
    definition_period = YEAR



from .congress.delauro import create_american_family_act_with_baby_bonus_reform
from .dc_kccatc import create_dc_kccatc_reform
from .winship import create_eitc_winship_reform
from .dc_tax_threshold_joint_ratio import (
    create_dc_tax_threshold_joint_ratio_reform,
)
from policyengine_core.reforms import Reform
import warnings


def create_structural_reforms_from_parameters(parameters, period):
    afa_reform = create_american_family_act_with_baby_bonus_reform(
        parameters, period
    )
    winship_reform = create_eitc_winship_reform(parameters, period)
    dc_kccatc_reform = create_dc_kccatc_reform(parameters, period)
    dc_tax_threshold_joint_ratio_reform = (
        create_dc_tax_threshold_joint_ratio_reform(parameters, period)
    )

    reforms = [
        afa_reform,
        winship_reform,
        dc_kccatc_reform,
        dc_tax_threshold_joint_ratio_reform,
    ]
    reforms = tuple(filter(lambda x: x is not None, reforms))

    class combined_reform(Reform):
        def apply(self):
            for reform in reforms:
                reform.apply(self)

    return combined_reform



import os
from fiscalsim_us import Microsimulation
from fiscalsim_us.data import CPS_2023
from fiscalsim_us.tools.taxsim.generate_taxsim_tests import TaxSim35
import numpy as np
import pytest
import pandas as pd
import platform

# Disable warnings
import warnings

warnings.filterwarnings("ignore")
warnings.simplefilter("ignore")

STATES = ["MD", "MA", "NY", "WA"]
DISTANCE = 100
MINIMUM_PERCENT_CLOSE = 0

if os.name != "nt":

    @pytest.fixture(scope="module")
    def taxsim():
        taxsim = TaxSim35()

        yield taxsim.generate_from_microsimulation(
            CPS_2023, 2022, None, True, False
        ).set_index("taxsim_taxsimid")

    @pytest.fixture(scope="module")
    def sim():
        yield Microsimulation()


@pytest.mark.skipif(True, reason="This test temporarily suspended.")
def test_federal_tax_against_taxsim(sim, taxsim):
    tax = sim.calc("income_tax")
    tax.index = sim.calc("tax_unit_id").values
    comparison_df = pd.DataFrame(index=sim.calc("tax_unit_id").values)
    comparison_df["fiscalsim_us"] = tax
    comparison_df["taxsim"] = taxsim.taxsim_fiitax
    relative_distance = np.absolute(
        comparison_df.fiscalsim_us - comparison_df.taxsim
    )
    percent_close = (relative_distance < DISTANCE).mean()
    assert percent_close > MINIMUM_PERCENT_CLOSE


@pytest.mark.skipif(True, reason="This test temporarily suspended.")
@pytest.mark.parametrize("state", STATES)
def test_state_income_tax_against_taxsim(state: str, sim, taxsim):
    in_state = sim.calc("tax_unit_state").values == state
    tax = sim.calc("state_income_tax")
    tax.index = sim.calc("tax_unit_id").values
    comparison_df = pd.DataFrame(
        dict(
            fiscalsim_us=tax,
            taxsim=taxsim.taxsim_siitax,
        ),
        index=sim.calc("tax_unit_id").values,
    )
    comparison_df = comparison_df[in_state]
    relative_distance = np.absolute(
        comparison_df.fiscalsim_us - comparison_df.taxsim
    )
    percent_close = (relative_distance < DISTANCE).mean()
    assert percent_close > MINIMUM_PERCENT_CLOSE



from fiscalsim_us.model_api import *


class va_adj_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Line 9 on form 760 https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/760-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        line3 = tax_unit("va_calc_line_3", period)
        line8 = tax_unit("va_calc_line_8", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values

        threshold = parameters(
            period
        ).gov.states.va.tax.income.va_adjusted_gross_income
        single = threshold.SINGLE
        joint = threshold.JOINT

        subtotal = line3 - line8

        tax_owed = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.SEPARATE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE),
            where(subtotal < single, 0, subtotal),
            where(
                filing_status == filing_statuses.JOINT,
                where(subtotal < joint, 0, subtotal),
                subtotal,
            ),
        )

        return tax_owed



from fiscalsim_us.model_api import *


class oh_retirement_income_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio Retirement Income Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://codes.ohio.gov/ohio-revised-code/section-5747.055"
    defined_for = "oh_retirement_income_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.retirement_income

        person = tax_unit.members
        is_spouse = person("is_tax_unit_spouse", period)
        is_head = person("is_tax_unit_head", period)
        pension_income = person("pension_income", period)
        has_not_taken_lump_sum_distribution = person(
            "oh_has_not_taken_oh_lump_sum_credits", period
        )
        head_or_spouse = is_head | is_spouse
        eligible_pension = (
            pension_income
            * has_not_taken_lump_sum_distribution
            * head_or_spouse
        )
        total_pension_income = tax_unit.sum(eligible_pension)

        return p.amount.calc(total_pension_income, right=True)



from fiscalsim_us.model_api import *


class or_income_after_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR income after additions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR
    adds = ["adjusted_gross_income", "or_income_additions"]



from fiscalsim_us.model_api import *


class il_nonrefundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL nonrefundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    adds = "gov.states.il.tax.income.credits.nonrefundable"



from fiscalsim_us.model_api import *


class ar_metabolic_disorder_expenses(Variable):
    "Line 4 of form AR1000TC"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Phenylketonuria Disorder expenses"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TC_Schedule_of_TaxCredits_andBusiness_IncentiveCredits.pdf"
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class ar_distribution_income(Variable):
    """Line 31 of form AR1000F
    Line 1 of form AR1000TD"""

    value_type = float
    entity = TaxUnit
    label = "Arkansas Lump Sum Distribution  Income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TD_LumpSumDistributionAveraging.pdf"
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class sey(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    unit = USD

    adds = ["self_employment_income", "farm_income", "k1bx14"]


class filer_sey(Variable):
    value_type = float
    entity = TaxUnit
    label = "sey for the tax unit (excluding dependents)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("sey", tax_unit, period)


class combined(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxes"
    documentation = "Total federal income and payroll tax liability."
    unit = USD

    adds = [
        # Tax unit level.
        "iitax",
        "additional_medicare_tax",
        # Person level
        "self_employment_medicare_tax",
        "self_employment_social_security_tax",
        "employee_medicare_tax",
        "employee_social_security_tax",
    ]


tax = variable_alias("tax", combined)


class filer_earned(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD

    def formula(tax_unit, period, parameters):
        return max_(0, tax_unit_non_dep_sum("earned", tax_unit, period))


class earned(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Earned income"
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD

    def formula(person, period, parameters):
        misc = parameters(period).gov.irs.ald.misc
        adjustment = (
            (1 - misc.self_emp_tax_adj)
            * misc.employer_share
            * person("self_employment_tax", period)
        )
        return person("earned_income", period) - adjustment


class othertaxes(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Other taxes: sum of niit, e09700, e09800 and e09900 (included in c09200)"
    unit = USD


class sep(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    default_value = 1
    documentation = (
        "2 when filing_status is 3 (married filing separately); otherwise 1"
    )


class surtax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c01000(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Limitation on capital losses"
    unit = USD

    def formula(tax_unit, period, parameters):
        return max_(
            (-3000.0 / tax_unit("sep", period)), tax_unit("c23650", period)
        )


tax_unit_net_capital_gains = variable_alias(
    "tax_unit_net_capital_gains", c01000
)


class c07240(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c07260(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c07300(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c07400(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c07600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class c08000(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class tax_unit_is_joint(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Joint-filing tax unit"
    documentation = "Whether this tax unit is a joint filer."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        return filing_status == filing_status.possible_values.JOINT


class care_deduction(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )


class ctc_new(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class charity_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Credit for charitable giving"
    unit = USD


class fstax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD


class benefit_value_total(Variable):
    value_type = float
    entity = TaxUnit
    label = "Total benefit value"
    definition_period = YEAR



from fiscalsim_us.model_api import *


class ar_other_limited_expenses(Variable):
    """
    Line 21 of 2022 AR3, Arkansas Itemized Deductions

    Other deductions include:
    Union or professional dues
    Tax return preparation fees
    Expenses for safety equipment
    Expenses of entertaining customers
    Tools and supplies
    Fees paid to employment agencies
    """

    value_type = float
    entity = TaxUnit
    label = "AR other limited expenses"
    unit = USD
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR3_ItemizedDeduction.pdf"
    definition_period = YEAR
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *
import numpy as np


class ar_low_income_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas low income tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_TaxTables.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        tax_rate = parameters(period).gov.states.ar.tax.income.rates.rates
        credit_rate = parameters(
            period
        ).gov.states.ar.tax.income.credits.low_income_credit.income_tax_credit_pct
        num_dependents = tax_unit("tax_unit_dependents", period)
        min_income = where(
            num_dependents < 2,
            parameters(
                period
            ).gov.states.ar.tax.income.credits.low_income_credit.low_dep_inc_tax_credit_min_income[
                filing_status
            ],
            parameters(
                period
            ).gov.states.ar.tax.income.credits.low_income_credit.high_dep_inc_tax_credit_min_income[
                filing_status
            ],
        )
        phaseout_rate = where(
            num_dependents < 2,
            parameters(
                period
            ).gov.states.ar.tax.income.credits.low_income_credit.low_dep_inc_tax_credit_phaseout_rate[
                filing_status
            ],
            parameters(
                period
            ).gov.states.ar.tax.income.credits.low_income_credit.high_dep_inc_tax_credit_phaseout_rate[
                filing_status
            ],
        )
        agi = tax_unit("ar_agi", period)

        def round_to_nearest_50(num):
            # Calculate the nearest multiple of 100
            nearest_multiple_of_100 = np.round(num / 100) * 100

            # Get the last two digits
            last_two_digits = num % 100

            # Determine the closest ending in "50"
            if np.isscalar(num):
                if last_two_digits <= 50:
                    rounded_income = nearest_multiple_of_100 + 50
                    return rounded_income
                else:
                    rounded_income = nearest_multiple_of_100 - 50
                    return rounded_income
            else:
                rounded_income = np.zeros_like(num)
                rounded_income[last_two_digits <= 50] = (
                    nearest_multiple_of_100[last_two_digits <= 50] + 50
                )
                rounded_income[last_two_digits > 50] = (
                    nearest_multiple_of_100[last_two_digits > 50] - 50
                )
                return rounded_income

        std_ded = tax_unit("ar_standard_deduction", period)
        itm_ded = tax_unit("ar_itemized_deductions", period)
        deduction = where(itm_ded > std_ded, itm_ded, std_ded)

        agi_less_ded = agi - deduction
        min_inc_less_ded = min_income - deduction

        rounded_income = round_to_nearest_50(agi_less_ded)
        rounded_min_income = round_to_nearest_50(min_inc_less_ded)

        # Calculate the tax liability on min_income
        min_tax_liability = tax_rate.calc(rounded_min_income)

        min_tax_liability = np.round(min_tax_liability)

        # Calculate the credit amount
        credit_amount = min_tax_liability * credit_rate

        # Calculate the phaseout reduction for each $100 over min_income
        excess_income = rounded_income - rounded_min_income
        phaseout_count = where(
            excess_income % 100 > 0,
            excess_income // 100 + 1,
            excess_income // 100,
        )
        phaseout_reduction = phaseout_count * phaseout_rate

        # Reduce the credit amount based on the phaseout reduction
        credit_amount -= phaseout_reduction

        # Ensure credit_amount does not go below 0 and that those who itemize or are married filing separately do not take the credit
        if np.isscalar(credit_amount):
            credit_amount = where(
                credit_amount < 0
                or std_ded < itm_ded
                or filing_status == "SEPARATE",
                0,
                credit_amount,
            )
        else:
            credit_amount[
                (
                    (credit_amount < 0)
                    | (std_ded < itm_ded)
                    | (filing_status == "SEPARATE")
                )
            ] = 0

        credit_amount = np.round(credit_amount, 0)

        return credit_amount



from fiscalsim_us.model_api import *


class state_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "state income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    adds = [
        "ar_income_tax_before_refundable_credits",
        "ca_income_tax_before_refundable_credits",
        "co_income_tax_before_refundable_credits",
        "dc_income_tax_before_refundable_credits",
        "ia_income_tax_before_refundable_credits",
        "il_total_tax",
        "in_income_tax_before_refundable_credits",
        "ks_income_tax_before_refundable_credits",
        "ky_income_tax_before_refundable_credits",
        "la_income_tax_before_refundable_credits",
        "ma_income_tax_before_refundable_credits",
        "md_income_tax_before_refundable_credits",
        "me_income_tax_before_refundable_credits",
        "mn_income_tax_before_refundable_credits",
        "mt_income_tax_before_refundable_credits",
        "mo_income_tax_before_refundable_credits",
        "nc_income_tax",  # NC has no refundable credits.
        "nd_income_tax_before_refundable_credits",
        "ne_income_tax_before_refundable_credits",
        "nh_income_tax_before_refundable_credits",
        "nj_income_tax_before_refundable_credits",
        "nm_income_tax_before_refundable_credits",
        "ny_income_tax_before_refundable_credits",
        "ok_income_tax_before_refundable_credits",
        "or_income_tax_before_refundable_credits",
        "pa_income_tax",  # PA has no refundable credits.
        "ri_income_tax_before_refundable_credits",
        "sc_income_tax_before_refundable_credits",
        "ut_income_tax_before_refundable_credits",
        "va_income_tax_before_refundable_credits",
        "vt_income_tax_before_refundable_credits",
        "wa_income_tax_before_refundable_credits",
        "wi_income_tax_before_refundable_credits",
    ]



from fiscalsim_us.model_api import *


class de_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware refundable EITC"
    unit = USD
    documentation = (
        "Refundable EITC credit reducing DE State income tax page 8."
    )
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
    defined_for = "de_claims_refundable_eitc"

    adds = ["de_refundable_eitc_if_claimed"]



from fiscalsim_us.model_api import *
from fiscalsim_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_income_deemed_from_ineligible_spouse(Variable):
    value_type = float
    entity = Person
    label = "SSI income (deemed from ineligible spouse)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1163"

    def formula(person, period, parameters):
        personal_earned_income = person("ssi_earned_income", period)
        personal_unearned_income = person("ssi_unearned_income", period)
        spousal_earned_income = person(
            "ssi_earned_income_deemed_from_ineligible_spouse", period
        )
        spousal_unearned_income = person(
            "ssi_unearned_income_deemed_from_ineligible_spouse", period
        )

        amount = parameters(period).gov.ssa.ssi.amount

        income_if_combined = _apply_ssi_exclusions(
            personal_earned_income + spousal_earned_income,
            personal_unearned_income + spousal_unearned_income,
            parameters,
            period,
        )

        income_if_not_combined = _apply_ssi_exclusions(
            personal_earned_income,
            personal_unearned_income,
            parameters,
            period,
        )

        spousal_deemed_income = income_if_combined - income_if_not_combined

        ssi = parameters(period).gov.ssa.ssi.amount
        person_rate = (
            person("is_ssi_ineligible_child", period)
            * (ssi.couple - ssi.individual)
            * MONTHS_IN_YEAR
        )

        return spousal_deemed_income * (spousal_deemed_income > person_rate)



from fiscalsim_us.model_api import *


class ms_self_employment_adjustment(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi self employment adjustment"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=13",  # Line 61
        "https://law.justia.com/codes/mississippi/2020/title-27/chapter-7/article-1/section-27-7-18/",  # 6(c)
    ]
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        self_employment_tax = add(tax_unit, period, ["self_employment_tax"])
        p = parameters(
            period
        ).gov.states.ms.tax.income.adjustments.self_employment
        return self_employment_tax * p.rate



from fiscalsim_us.model_api import *


class prior_year_local_income_tax_paid(Variable):
    value_type = float
    entity = Person
    label = "Prior year local income taxes paid"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class ne_cdcc_refundable(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE refundable cdcc"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_2441n.pdf"
        "https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/Form_2441N_Ne_Child_and_Dependent_Care_Expenses_8-618-2022_final_2.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ne.tax.income.credits
        # determine AGI eligibility
        us_agi = tax_unit("adjusted_gross_income", period)
        agi_eligible = us_agi <= p.cdcc.agi_threshold
        # determine NE refundable cdcc amount
        us_cdcc = tax_unit("cdcc", period)
        ne_cdcc = us_cdcc * p.cdcc.refundable.fraction.calc(us_agi)
        return agi_eligible * ne_cdcc



from fiscalsim_us.model_api import *


class mmn_prev_year_state_refund(Variable):
    """
    Line 6 of Minnesota 2023 Individual Income Tax return from M1. This is the
    State income refund from a previous year that was reported in line 1 of
    Schedule 1 (Form 1040) of the federal return.
    """

    value_type = float
    entity = TaxUnit
    label = "MN other subtrations from income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-12/m1-23.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-02/m1-inst-23.pdf"
        "https://www.irs.gov/pub/irs-pdf/f1040s1.pdf"
    )
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class va_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA interim step to get to virginia taxable income, line 11 on form 760 https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values

        standard_deduction = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            | (filing_status == filing_statuses.SEPARATE),
            parameters.gov.states.va.tax.income.va_standard_deduction.SINGLE,
            where(
                filing_status == filing_statuses.JOINT,
                parameters.gov.states.va.tax.income.va_standard_deduction.JOINT,
                parameters.gov.states.va.tax.income.va_standard_deduction.SINGLE,
            ),
        )

        return standard_deduction



from fiscalsim_us.model_api import *


class lifeline(Variable):
    value_type = float
    entity = SPMUnit
    label = "Lifeline"
    documentation = "Amount of Lifeline phone and broadband benefit"
    definition_period = YEAR
    unit = USD
    reference = "https://www.law.cornell.edu/cfr/text/47/54.403"

    def formula(spm_unit, period, parameters):
        # NB: Only one Lifeline benefit is available per SPM unit, per:
        # https://www.law.cornell.edu/cfr/text/47/54.409#c
        amounts = parameters(period).gov.fcc.lifeline.amount
        household = spm_unit.household
        is_rural_tribal = and_(
            household, period, ["is_rural", "is_on_tribal_land"]
        )
        max_amount = (
            amounts.standard
            + is_rural_tribal * amounts.rural_tribal_supplement
        ) * MONTHS_IN_YEAR
        phone_broadband_cost = add(
            spm_unit, period, ["phone_cost", "broadband_cost"]
        )
        amount_if_eligible = min_(phone_broadband_cost, max_amount)
        eligible = spm_unit("is_lifeline_eligible", period)
        return eligible * amount_if_eligible



from fiscalsim_us.model_api import *


class mt_aged_exemption_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of Montana aged exemptions"
    definition_period = YEAR
    reference = "https://regulations.justia.com/states/montana/department-42/chapter-42-15/subchapter-42-15-4/rule-42-15-402/"
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.exemptions
        aged_head = (tax_unit("age_head", period) >= p.age_threshold).astype(
            int
        )
        aged_spouse = (
            tax_unit("age_spouse", period) >= p.age_threshold
        ).astype(int)
        return aged_head + aged_spouse



from fiscalsim_us.model_api import *


class oh_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://codes.ohio.gov/ohio-revised-code/section-5747.055",
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=20",
    )
    defined_for = StateCode.OH

    adds = [
        "adjusted_gross_income",
        "oh_bonus_depreciation_add_back",
        "oh_other_add_backs",
    ]

    subtracts = [
        "oh_section_179_expense_add_back",
        "qualified_business_income_deduction",
        "tax_unit_taxable_social_security",
        "dividend_income",
        # "above_the_line_deductions" #only need line 8z
        "oh_uniformed_services_retirement_income_deduction",
        "investment_in_529_plan",
        "spm_unit_pell_grant",
        "educator_expense",
        "disability_benefits",
    ]



from fiscalsim_us.model_api import *
from policyengine_core.simulations import Simulation


class meets_ssi_resource_test(Variable):
    value_type = bool
    entity = Person
    label = "Meets SSI resource test"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        simulation: Simulation = person.simulation
        if hasattr(simulation, "dataset"):
            pass_rate = parameters(
                period
            ).gov.ssa.ssi.eligibility.resources.pass_rate
            return random(person) < pass_rate
        joint_claim = person("ssi_claim_is_joint", period)
        personal_resources = person("ssi_countable_resources", period)
        countable_resources = where(
            joint_claim,
            person.marital_unit.sum(personal_resources),
            personal_resources,
        )
        ssi = parameters(period).gov.ssa.ssi
        resource_limits = ssi.eligibility.resources.limit
        resource_limit = where(
            joint_claim, resource_limits.couple, resource_limits.individual
        )
        return countable_resources <= resource_limit



from fiscalsim_us.model_api import *


class mt_federal_income_tax_deduction(Variable):
    """
    Line 4 on itemized deductions schedule
    Includes the following
    *Federal income tax withheld
    *Federal estimated tax payments
    *Prior year federal income taxes
    *Other back year federal income taxes
    Add each of those together but not more then $5,000 if you are single,
        head of household, or married filing sperarately; or $10,000 if you
        are married filing jointly
    """

    value_type = float
    entity = TaxUnit
    label = "Montana federal tax paid/withheld in prior year"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class xtot(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "Filing unit exemptions"
    documentation = "Total number of exemptions for filing unit"

    def formula(tax_unit, period, parameters):
        return tax_unit.nb_persons()



from fiscalsim_us.model_api import *


class tanf_max_amount(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF maximum benefit"
    documentation = "The maximum benefit amount a family could receive from Temporary Assistance for Needy Families given their state and family size."
    unit = USD

    def formula_2022(spm_unit, period, parameters):
        household_size = spm_unit("spm_unit_size", period).astype(str)
        state = spm_unit.household("state_code_str", period)
        max_amount = parameters(period).gov.hhs.tanf.cash.amount.max
        return max_amount[state][household_size] * MONTHS_IN_YEAR



from fiscalsim_us.model_api import *


class il_base_income_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL base income subtractions"
    unit = USD
    definition_period = YEAR

    adds = "gov.states.il.tax.income.base.subtractions"



from fiscalsim_us.model_api import *


class pha_payment_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD payment standard"
    unit = USD
    documentation = "Payment standard for HUD programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/982.503"



from fiscalsim_us.model_api import *


class mn_unreimbursed_employee_deduction(Variable):
    """
    Line 20 of 2023 M1SA, Minnesota Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "MN unreimbursed employee expenses deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        fed_agi = tax_unit("adjusted_gross_income", period)
        mult = parameters(
            period
        ).gov.states.mn.tax.income.deductions.unreimbursed_employee_mult
        expense = tax_unit("mn_unreimbursed_employee_expenses", period)

        return max_(expense - fed_agi * mult, 0)
